import json
import logging
import sys

from logdecorator import log_on_start, log_on_end, log_on_error, log_exception

import csv
from typing import Dict, List

import io
import boto3
import requests
from requests.adapters import HTTPAdapter
from urllib3 import Retry

##
# @file worker.py
# Cloud evaluation worker.
#
# Its main function is to run against a model in cloud environment, it runs on docker in parallel to the server.
#
# Basically it simulates access to the server in a similar way as in production.
#
# It performs the following steps:
# - Reads dataset from S3
# - Invoke model on Gunicorn server - /parse /predict
# - Receives results from server.
# - Write results to S3
#
# Receives two parameters from CLI args:
# @param input_dataset_filename - Input dataset filename
# @param output_dataset_filename - Output results filename

service_client = None

endpoint_name = None


@log_on_start(logging.INFO, "loading dataset {input_path:s}...")
@log_on_error(logging.ERROR, "Error on loading dataset {input_path:s}: {e!r}",
              on_exceptions=IOError,
              reraise=True)
@log_on_end(logging.INFO, "Downloading {input_path:s} finished successfully")
def load_csv(input_path):
    """! Loads a csv file from S3 bucket

        Args:
            input_path: Input file path
        Returns:
            rows: List of dataset rows.

    """
    with download_file_from_s3(input_path) as csvfile:
    # with open(input_path, newline='') as csvfile:
        input_reader = csv.DictReader(csvfile, delimiter=',')
        rows = [row for row in input_reader]
        return rows


@log_on_start(logging.DEBUG, "invoke_model on {row}...")
def invoke_model(row):
    """! Sends a row of data (from the input dataset) to the model at the specified endpoint.

        This method prepares a header and payload based on a row of data from the dataset, sends it to the server
        and returns prediction result.

        Args:
            row: Row of data taken from the input dataset

        Returns:
            model_prediction - Prediction result in json format.
    """
    request_payload = service_client.get_request_payload(row)
    headers = service_client.get_request_headers(row)

    target_host = "127.0.0.1"
    target_port = 8080
    url = f"http://{target_host}:{target_port}/{endpoint_name}"

    retries = Retry(total=5, backoff_factor=1, read=5, status=5,  # allowed_methods=['GET', 'POST'],
                    status_forcelist=[502, 503, 504, 400])
    s = requests.Session()
    s.mount('http://', HTTPAdapter(max_retries=retries))

    try:
        response = s.post(
            url=url,
            json=request_payload,
            headers=headers
        )
    except requests.ConnectionError as ce:
        raise Exception(
            f"Error connecting to service at {url}. Is the service up and running, with the correct port?") from ce
    except Exception as e:
        raise Exception(f"Error call service") from e

    response_json = response.json()
    if "error" in response_json:
        raise Exception("Error in model service")

    model_prediction = service_client.extract_model_predictions_from_response(response_json)

    return model_prediction


@log_on_start(logging.INFO, "Starting work on dataset...")
@log_on_error(logging.ERROR, "Error invoking model on dataset: {e!r}",
              on_exceptions=IOError,
              reraise=True)
@log_on_end(logging.INFO, "Completed working on dataset")
def work_on(rows):
    """! The main cloud evaluation method.

        Iterates dataset rows, invokes model and evaluates each row.

        Args:
            rows: Input dataset

        Returns:
            rows - Includes prediction and evaluation results.
    """
    for row in rows:
        # Call model and get prediction
        row['prediction'] = json.dumps(invoke_model(row))
        # Run evaluation and get confusion matrix
        row['evaluation'] = service_client.evaluate_model_results(row)
        # Format the output structure
        service_client.set_format_output(row)
    return rows


def write_to_s3(csv_content, bucket_name, file_path):
    """ Uploads file to S3

    Args:
        csv_content: Csv file in bytes array
        bucket_name: Bucket name
        file_path: File output path

    """
    client = boto3.client('s3')
    response = client.put_object(
        Body=csv_content,
        Bucket=bucket_name,
        Key=file_path
    )


@log_on_start(logging.INFO, "writing dataset {output_path:s}...")
@log_on_error(logging.ERROR, "Error on write dataset {output_path:s}: {e!r}",
              on_exceptions=IOError,
              reraise=True)
@log_on_end(logging.INFO, "Writing {output_path:s} finished successfully")
def write_output(output: List[Dict], output_path):
    """! Writes prediction and evaluation results to S3 bucket.

            Initially prepare results in csv format and calls write_to_s3() to write results.

            Args:
                output: List of results.
                output_path: Path to save results in S3.

    """
    bucket_name, file_path = parse_s3_url(output_path)
    with io.StringIO() as csvfile:
    # with open(output_path, 'w', newline='') as csvfile:
        fieldnames = list(output[0].keys())
        writer = csv.DictWriter(csvfile, fieldnames=fieldnames, quotechar='"')

        writer.writeheader()
        for row in output:
            writer.writerow(row)

        csv_content = bytes(csvfile.getvalue(), encoding='utf-8')

        write_to_s3(csv_content, bucket_name, file_path)


def main(input_path: str, output_path: str):
    """! Cloud evaluation main method.

        Does the following:
        - Loads csv input dataset
        - Runs model on input dataset
        - Write results to output_path

        Args:
            input_path: Path to input dataset
            output_path: Path to results

    """
    # Load CSV
    rows = load_csv(input_path)
    # Call model and evaluate on each row
    output = work_on(rows)
    # Write output file
    write_output(output, output_path)


def download_file_from_s3(source_s3_url: str):
    """! Downloads file from S3 using boto3 package.

        Args:
            source_s3_url: S3 Url
        Returns:
            file_stream - Downloaded file
    """
    bucket_name, filepath = parse_s3_url(source_s3_url)
    s3 = boto3.client('s3', region_name='us-east-2')
    res = s3.get_object(Bucket=bucket_name, Key=filepath)['Body'].read()
    file_stream = io.StringIO(res.decode('utf-8'))
    return file_stream


def parse_s3_url(s3_url: str):
    """! Parse S3 url to bucket name and file path

        Args:
            s3_url: Url

        Returns:
            bucket_name - Bucket name
            filepath - File path
    """
    bucket_name, filepath = s3_url.split('//')[1].split('/', maxsplit=1)
    return bucket_name, filepath


# if __name__ == '__main__':
#     # input = "./test/dummy_dataset.csv"
#     input = "s3://ds-cloud-eval-storage/datasets/dummy_dataset.csv"
#     # output = "./test/dummy_output.csv"
#     output = "s3://ds-cloud-eval-storage/predictions/dummy_output.csv"
#     main(input, output)

if __name__ == '__main__':
    from cloud_eval.cloud_eval_client import CloudEvalClient

    ## CLI argv parsed:
    ## @param input_dataset_filename - Input dataset filename
    ## @param output_dataset_filename - Output results filename
    input_dataset_filename, output_dataset_filename = sys.argv[1:1 + 2]
    ## Instance of the CloudEvalClient
    service_client = CloudEvalClient()
    ## 'predict' or 'parse' - depends on a specific project.
    endpoint_name = service_client.get_endpoint_name()
    main(input_dataset_filename, output_dataset_filename)



#!/usr/bin/env python
# coding: utf-8

"""! @brief ZIDS_SharedArtifacts base class for the SharedArtifacts class."""

from typing import List, Any
import sys
import pathlib
import os
import json
import csv
import pickle
from dsframework.base.common import functions

class ZIDS_SharedArtifacts():
    """! ZIDS_SharedArtifacts is the base class for the SharedArtifacts class."""

    def __init__(self) -> None:
        """! ZIDS_SharedArtifacts initializer.
        Loads config, vocabs. and artifacts.
        """
        file_path = sys.modules[self.__class__.__module__].__file__
        curr_path = file_path[:file_path.rfind("pipeline")]
        ## Path for the framework.
        self.base_dir = curr_path
        self.load_config_json()
        self.load_vocabs()
        self.load_artifacts()

    def load_config_json(self):
        """! Loads config and sets its content to class attributes.
        It enables adding additional required settings to the config file (located here: config/config.json),
        which will be auto-loaded and be accessible via the SharedArtifacts class.
        """
        if os.path.exists(self.base_dir + 'config/config.json'):
            data = functions.load_json(self.base_dir + 'config/config.json')
            for item in data:
                setattr(self, item, data[item])

    def load_vocabs(self):
        """! Loads vocabulary from the specified path in config.json to class attributes,
        accessible via SharedArtifacts class. file can be in any format, we have the following ready to use:
        json, csv, pickle, tensorflow, key_value and key_value_int.
        To add additional formats override the function extend_load_file_type for example:

        @verbatim
        def extend_load_file_type(self, file_type, path, absolute_path, name):
            if absolute_path:
                if file_type == 'your-file-type':
                    with open(absolute_path) as json_file:
                        setattr(self, name, json.load(json_file))
        @endverbatim

        Setting the config.json file, example:
        @verbatim
        "vocabs": [{"name":"my_vocabs" ,"path": "pipeline/artifacts/vocabs/example_vocabs.json", "type": "json"}]
        @endverbatim
        """
        if hasattr(self, 'vocabs'):
            for item in self.vocabs:
                self.load_file(item)

    def load_artifacts(self):
        """! Load artifacts from the specified path in config.json to local attributes,
        accessible via SharedArtifacts class.
        Data file can be in any format, we have the following formats implemented:
        json, csv, pickle, tensorflow, key_value and key_value_int.
        To add additional formats override the function extend_load_file_type for example:
        """
        ##
        # @code{.py}
        # def extend_load_file_type(self, file_type, path, absolute_path, name):
        #     if absolute_path:
        #         if file_type == 'your-file-type':
        #             with open(absolute_path) as json_file:
        #                 setattr(self, name, json.load(json_file))
        # @endcode
        #
        # Setting the config.json file, example:
        # @verbatim
        # "artifacts": [{"name": "my_artifacts", "path": "pipeline/artifacts/models/example_models.json", "type": "json"}]
        # @endverbatim

        if hasattr(self, 'artifacts'):
            for item in self.artifacts:
                self.load_file(item)

    def load_file(self, item):
        """! Loads a file, this method is used by load_vocabs and load_artifacts methods.

        Args:
            item: (dict) with the following keys: {'name': '', 'path': '', 'type': ''}
                name: assign any name,
                path: path to data file,
                type: file format, format supported: json, csv, pickle, tensorflow, key_value and key_value_int
                    additional formats can be added by overriding extend_load_file_type method.
        """
        file_type = item['type']
        path = item['path']
        name = item['name']
        if path:
            absolute_path = self.base_dir + '/' + path
            if file_type == 'json':
                self.load_json(absolute_path, name)
            elif file_type == 'csv':
                self.load_csv(absolute_path, name)
            elif file_type == 'pickle':
                self.load_pickle(absolute_path, name)
            elif file_type == 'tensorflow':
                self.load_tensorflow(absolute_path, name)
            elif file_type == 'key_value':
                self.load_file_to_dict(absolute_path, name, str)
            elif file_type == 'key_value_int':
                self.load_file_to_dict(absolute_path, name, int)
            else:
                self.extend_load_file_type(file_type, path, absolute_path, name)

    def extend_load_file_type(self, file_type, path, absolute_path, name):
        """! Use this method to add new file types, that are currently not supported.

        Override method in the generatedProjectNameSharedArtifacts class to support new file format.
        """

        ##
        # For example:
        # @code{.py}
        # def extend_load_file_type(self, file_type, path, absolute_path, name):
        #     if absolute_path:
        #         if file_type == 'your-file-type':
        #             with open(absolute_path) as json_file:
        #                 setattr(self, name, json.load(json_file))
        # @endcode

        pass

    def load_json(self, path, name):
        """! Loads json file

        Args:
            path: Absolute path to data file.
            name: File name.
        """
        with open(path) as json_file:
            setattr(self, name, json.load(json_file))

    def load_csv(self, path, name):
        """! Loads csv file

        Args:
            path: Absolute path to data file.
            name: File name.
        """
        with open(path) as csv_file:
            setattr(self, name, functions.flatten_list(list(csv.reader(csv_file))))

    def load_pickle(self, path, name):
        """! Loads pickle file

        Args:
            path: Absolute path to data file.
            name: File name.
        """
        with open(path, 'rb') as pickle_file:
            setattr(self, name, pickle.load(pickle_file))

    def load_tensorflow(self, path, name):
        """! Loads tensorflow file.

        Important:
            This method is not implemented, please override to use its functionality.

        Args:
            path: Absolute path to data file.
            name: File name.

        Raises:
            NotImplementedError
        """
        ##
        # Implementation example:
        # @code
        # with open(path) as tensorflow_file:
        #     setattr(self, name, tf.keras.models.load_model(tensorflow_file))
        # @endcode

        raise NotImplementedError


    def load_file_to_dict(self, path, name, val_type):
        """! Loads file to dictionary.

        File structure needs to be a comma seperated key and value, one per line:

        key, value
        key, value
        .
        .
        .

        Args:
            path: Absolute path to data file.
            name: File name.
            val_type: datatype ie. str, int
        """
        dictionary = functions.load_file_to_dict(path, value_type=val_type)
        setattr(self, name, dictionary)

    ##
    # @cond
    def get(self, key, default=None):
        if key in self.__dict__:
            return self.__dict__[key]
        return default

    def set(self, key, val, default=None):
        if key in self.__dict__:
            return self.__dict__.update({key: val})
        return default
    ##
    # @endcond



# !/usr/bin/env python
# coding: utf-8

from typing import List, Any

from dsframework.base.pipeline.preprocessor import ZIDS_Preprocessor
from ..schema.inputs import generatedProjectNameInputs
from ..schema.outputs import generatedProjectNameOutputs
from ..artifacts.shared_artifacts import generatedProjectNameSharedArtifacts

##
# @file
# @brief generatedClass (Preprocess) class, implements ZIDS_Preprocessor base class.
class generatedClass(ZIDS_Preprocessor):
    """! generatedClass (Preprocessor) class implements ZIDS_Preprocessor base class.

    First step of the pipeline, its main goal is to format the dataset format to the model requirements.

    Its base class ZIDS_Preprocessor declares phases based on UVM (Universal Verification Methodology) and by this it
    gives us a structured way to work in each one of the main components by overriding and implementing those phases
    in this class.

    Important note:
    Those methods are divided into two groups, the ones that run from the ZIDS_Component.__init__() and
    those that run from ZIDS_Component.execute() method, use them based on the required execution order.

    In the __init__() we have the following called:
    - build
    - config
    - config_from_json
    - connect

    In the execute(), we call the following:
    - reset
    - pre_run
    - run
    - post_run
    - evaluate
    """

    def __init__(self, artifacts: generatedProjectNameSharedArtifacts = None):
        """! generatedClass (Preprocess) initializer

        Args:
            artifacts(generatedProjectNameSharedArtifacts): Shared artifacts instance.
        """

        ##
        # @hidecallgraph @hidecallergraph
        super().__init__(artifacts)

    def normalize_input(self, **kwargs: Any) -> generatedProjectNameInputs:
        """! Converts dataset to generatedProjectNameInputs

        Args:
            **kwargs: Loaded dataset.
        Returns:
            The loaded dataset in generatedProjectNameInputs datatype.
        """
        return generatedProjectNameInputs(**kwargs)

    def config(self):
        """! Config method called from ZIDS_Component.__init__()

        Method not implemented, ready for your implementation, see more information in class description.
        """
        pass

    def connect(self):
        """! Connect method, called from ZIDS_Component.__init__().

        Method not implemented, ready for your implementation, see more information in class description.
        """

        ##
        #@hidecallgraph @hidecallergraph

        super().connect()

    def reset(self, text: str = '', hints: List[Any] = []):
        """! Reset data members, called from ZIDS_Component.execute() method.

        Method not implemented, see more information in class description.

        Args:
            text: str
            hints: List[Any]
        """

        ##
        # For example:
        # @code {.py}
        # def reset(self, text: str, raw_input: Any):
        #
        #     self.text = ""
        #     self.hints = raw_input.hints
        #
        # @endcode

        pass

    def get_from_regex(self):
        """! Get a predefined key from common/regex_handler.py

        Method not implemented, ready for your implementation.
        """
        ##
        # For example:
        # @code{.py}
        # from dsframework.base.common import RegexHandler
        #
        # def get_from_regex(self):
        #    return RegexHandler.url.findall("www.zoominfo.com")
        # @endcode

        pass

    def preprocess(self, raw_input: Any):
        """! Implement method to return a list of predictable objects in generatedProjectNameInputs datatype.

        Args:
            raw_input: Loaded dataset.

        Returns:
            List[ZIDS_Predictable] - Not implemented yet.

        Raises:
            NotImplementedError
        """

        ##
        # For example, minimum return as-is:
        # @code{.py}
        # def preprocess(self, raw_input: Any):
        #
        #     return raw_input
        # @endcode

        raise NotImplementedError



import json
from typing import List, Optional

from pydantic import validator, Extra
from pydantic.main import BaseModel


class TestOutputResponse(BaseModel):
    truth_id: int
    truth_dataset_id: int
    model_type_id: int
    raw_id: int = -1
    pred: str
    target: str
    text: str = ''

    @validator('pred')
    def pred_must_be_valid_json(cls, v):
        try:
            json.loads(v)
        except:
            raise ValueError('pred must be a valid json string')
        return v

    @validator('target')
    def target_must_be_valid_json(cls, v):
        try:
            json.loads(v)
        except:
            raise ValueError('target must be a valid json string')
        return v

    class Config:
        extra = Extra.allow



from typing import List, Any
from dsframework.base.common.component import ZIDS_Component

##
# @file
# @brief Predictor class, this is where the model goes.

class generatedClass(ZIDS_Component):
    """! generatedClass class (Predictor) implements ZIDS_Component base class.

    The second step of the pipeline and its main goal is to feed a dataset to a model.

    All the action happens in this step, model sits in this class.
    """

    def __init__(self, artifacts=None) -> None:
        """generatedClass class (Postprocessor) initializer

        Args:
            artifacts: Shared artifacts instance.
        """

        ##
        # @hidecallgraph @hidecallergraph
        super().__init__(artifacts)

    def execute(self, predictables: List[Any], **kwargs) -> List[Any]:
        """! generatedClass.execute (Predictor), Feed dataset to the model in this method, implement here.


        Args:
            predictables:List[Any]: List of predictable objects received from Preprocess.
            **kwargs: Dictionary with additional variables, if necessary.
        Returns:
            List[ZIDS_Predictable]: List of predictable objects.
        """

        ##
        # Implementation example:
        # @code{.py}
        #     def execute(self, predictables: List[Any], **kwargs) -> List[Any]:
        #         features = [list(vars(p.features).values()) for p in predictables]
        #
        #         results = self.artifacts.rf_model.predict_proba(features)
        #
        #         idx = 0
        #         for p in predictables:
        #             p.prob = results[idx][-1]
        #             idx += 1
        #         return predictables
        # @endcode

        ##
        # another example:
        # @code
        #     def execute(self, predictables: List[Any], **kwargs) -> List[Any]:
        #         df = pd.DataFrame([predictables.dict()])
        #         df = self.text_data_preparation(df)
        #         data = self.encode_batch(df['text'].tolist())
        #
        #         bert_output = self.artifacts.model(data)
        #
        #         probas = list(tf.nn.softmax(bert_output.logits).numpy())
        #         return probas
        # @endcode

        return predictables



from dsframework.pipeline import LabelizerBase
from dsframework.config import ZIDSConfig
from dsframework.shared import SharedArtifacts


class generatedClass(LabelizerBase):
    def __init__(self, name: str = "generatedClassName", model_path: str = '', use_dummy_model=False,
                 threshold: float = 0.5, shared_artifacts: SharedArtifacts = None, config: ZIDSConfig = None, **kwargs):
        super(generatedClass, self).__init__(name=name, model_path=model_path, use_dummy_model=use_dummy_model,
                               threshold=threshold, shared_artifacts=shared_artifacts, config=config, **kwargs)

    def __call__(self, **kwargs) -> LabelizerBase:
        raise NotImplementedError

    def init_predictor(self, model_path: str = '', use_dummy_model=False, threshold: float = 0.5) -> None:
        """
        init predictor of model (Catboost model)
        :param model_path: path to model
        :param use_dummy_model: use dummy model or not
        :param threshold: threshold
        """
        raise NotImplementedError

    def init_forcer(self, config: ZIDSConfig, shared_objects: SharedArtifacts):
        """
        init forcer
        """
        raise NotImplementedError



from cli.tester.server.main_template import generatedClass
from cli.tester.server.pool_template import generatedClass
from cli.tester.server.server_template_old import generatedClass
from cli.tester.server.test_server_post_template import generatedClass




"""! @brief Skeleton for implementing a specific stage to be used in a DataProc Workflow template."""
import sys
from typing import Any
from json import loads as json_loads

from dsframework.base.batch.dag_base import ZIDS_Dag


##
# @file
# @brief Stage main class, implements ZIDS_Stage base class.
class generatedDagName(ZIDS_Dag):
    """! Stage class

    Implement a stage that will later be converted to an executable job in a specific workflow.
    """

    def __init__(self, dag_config):
        """! The Stage class (generatedStageName) initializer.
        Base class will load basic configuration parameters, additional fields should be added here

        Args:
            dag_config : Configuration dictionary, loaded from configuration file.
        """

        ##
        # @hidecallgraph @hidecallergraph
        super().__init__(dag_config)

    def create_dag(self, **kwargs: Any):
        """! ZIDS_Stage main function.
        This function is the "entrypoint" for the stage, this will run when the job is executed.

            Args:
                **kwargs : Whatever is needed for the stage to run properly.
        """
        raise NotImplementedError


if __name__ == "__main__":
    """! Executes the dag by instantiating it and calling the main function.
    Set up argument condition according to the usage of the written stage
    
        Args:
            System argument 1 - Configuration file
    """
    if sys.argv and len(sys.argv) > 1:
        config = json_loads(sys.argv[1])
        dag = generatedStageName(config)
        dag.create_dag()
    else:
        print(f"project configuration not provided, Can't run dag")



import subprocess
import os
import sys
from json import dumps as json_dumps
from json import load as json_load

from google.cloud.dataproc_v1 import ClusterControllerClient
from google.cloud.dataproc_v1 import WorkflowTemplatePlacement
from google.cloud.dataproc_v1.types import OrderedJob
from google.cloud.dataproc_v1.types import PySparkJob
from google.cloud.dataproc_v1.services.workflow_template_service.async_client import WorkflowTemplateServiceClient
from google.cloud.dataproc_v1.types.workflow_templates import WorkflowTemplate


##
# @file
# @brief Defines workflow base class.
class ZIDS_Workflow():
    def __init__(self, pip_packages):
        """! ZIDS_Workflow initializer.
        Loads config file and initializes parameters.

            ARGS:
                 pip_packages: A list of pip packages read from requirements file.
        """

        file_path = sys.modules[self.__class__.__module__].__file__
        curr_path = file_path[:file_path.rfind("batch_files")]

        with open(curr_path + '/batch_files/batch_config.json') as batch_cfg_file:
            self.config = json_load(batch_cfg_file)

        self.user_email = ''
        env_type = os.environ.get('SPRING_PROFILES_ACTIVE')
        if (not env_type == 'production') and (not env_type == 'staging'):
            command = 'gcloud config get-value account'
            user_email = subprocess.check_output(command, shell=True).decode(sys.stdout.encoding)
            self.user_email = user_email.split("@")[0]

        if self.user_email:
            self.config['user_email'] = self.user_email

        self.default_args = [json_dumps(self.config), "start_date_placeholder", "end_date_placeholder"]

        self.region = self.config['region']
        self.zone = self.region + '-b'
        self.project_id = self.config['project_id']
        self.template_id = self.config['template_id']
        self.bucket_name = self.config['bucket_name']
        self.project_name = self.config['project_name']
        self.whl_file_name = self.project_name + '.zip'
        self.autoscaling_policy_id = self.project_name + '-auto-scaling-policy'

        self.unique_iteration_id = self.config['unique_iteration_id']
        self.unique_template_id = self.config['unique_template_id']
        if self.unique_template_id:
            self.template_id = self.template_id + '_' + self.unique_template_id

        if self.unique_iteration_id:
            self.folder_path = self.project_name + '/' + self.user_email + '/unique_iteration_id_' + self.unique_iteration_id
        else:
            self.folder_path = self.project_name + '/' + self.user_email + '/main'

        self.bucket_path = f'gs://{self.bucket_name}/{self.folder_path}'
        self.project_path = 'projects/{project_id}/regions/{region}'.format(project_id=self.project_id,
                                                                            region=self.region)

        if self.config['cluster_conf']['managed_cluster']:
            self.config['cluster_conf']['managed_cluster']['config']['gce_cluster_config']['metadata'][
                'PIP_PACKAGES'] = pip_packages

    def check_cluster(self):
        """! ZIDS_Workflow get_cluster.
        Tries to locate existing cluster that should be set up by this point.
        """
        clusterClient = ClusterControllerClient(
            client_options={"api_endpoint": f"{self.region}-dataproc.googleapis.com:443"}
        )
        cluster_name = self.config['cluster_conf']['managed_cluster']['cluster_name']
        try:
            exist_cluster = clusterClient.get_cluster(
                project_id=self.project_id,
                region=self.region,
                cluster_name=cluster_name,
            )
            if exist_cluster:
                del self.config['cluster_conf']['managed_cluster']
                self.config['cluster_conf']['cluster_selector'] = {
                    "cluster_labels": {
                        "goog-dataproc-cluster-name": cluster_name
                    }
                }
        except Exception as e:
            print(f'error getting cluster: {e}')

    def create_template_structure(self) -> (WorkflowTemplate, WorkflowTemplateServiceClient):
        """! ZIDS_Workflow create_template_structure.
        Tries to create a workflow template, containing all available stages.

            Returns:
                WorkflowTemplate() - Created template with basic arguments
                WorkflowTemplateServiceClient() - Service client for template handling
        """
        cluster_config = WorkflowTemplatePlacement(self.config['cluster_conf'])

        client = WorkflowTemplateServiceClient(
            client_options={"api_endpoint": f"{self.region}-dataproc.googleapis.com:443"}
        )
        # if results not exist
        template = WorkflowTemplate()
        template.id = self.template_id
        template.name = self.project_path + '/workflowTemplates/' + template.id
        template.placement = cluster_config
        return client, template

    def create_ordered_job(self, stage_dir) -> OrderedJob:
        """! ZIDS_Workflow create_ordered_job.
        Create an ordered job from a stage directory

            ARGS:
                stage_dir: Stage directory, should hold the main.py file and configuration
                           example: batch_files/stages/stage_write_page_list_to_big_query
            RETURNS:
                An ordered job
        """

        # Assume that the last part of the path is the stage name
        stage_name = stage_dir.rsplit('/', 1)[-1]

        created_stage = OrderedJob()
        pyJob = PySparkJob()
        pyJob.main_python_file_uri = f'{self.bucket_path}/{stage_dir}/main.py'
        pyJob.archive_uris = [f'{self.bucket_path}/dist/{self.whl_file_name}']
        pyJob.python_file_uris = [f'{self.bucket_path}/dist/{self.whl_file_name}']
        pyJob.args = self.default_args + []
        pyJob.jar_file_uris = self.default_jars + []
        created_stage.step_id = f'{self.project_name}-stage_{stage_name}'
        created_stage.pyspark_job = pyJob
        return created_stage

    @staticmethod
    def check_template(serv_client, template):
        """! Check for an existing template. If found, delete it.
            ARGS:
                serv_client: WorkflowTemplateServiceClient()
                template: WorkflowTemplate()
        """

        exist_template = None

        # Getting template if it exists
        try:
            exist_template = serv_client.get_workflow_template(
                name=template.name,
            )
            print(f'template {template.id} retrieved successfully')
        except Exception as e:
            print(f'Error getting template: {e}')

        # Deleting template if it exists
        if exist_template:
            try:
                serv_client.delete_workflow_template(
                    name=template.name,
                )
                print(f'template {template.id} deleted successfully')
            except Exception as e:
                print(f'Error deleting template: {e}')

    def create_template(self, serv_client, template):
        """! Create the complete template on the remote location.
            ARGS:
                serv_client: WorkflowTemplateServiceClient()
                template: WorkflowTemplate()
        """
        try:
            result = serv_client.create_workflow_template(
                parent=self.project_path,
                template=template
            )
            print(f'template {template.id} created successfully')
        except Exception as e:
            print(f'Error creating template: {e}')

    @staticmethod
    def instantiate_template(serv_client, template):
        """! Instantiate the template into a working workflow, overriding DAG
        Do this only on dev mode
            ARGS:
                serv_client: WorkflowTemplateServiceClient()
                template: WorkflowTemplate()
        """
        try:
            serv_client.instantiate_workflow_template(
                name=template.name,
                parameters={"START_DATE": '2022-02-23T00:00:00+00:00', "END_DATE": '2022-02-24T00:00:00+00:00'}
            )
            print(f'template {template.id} instantiated successfully')
        except Exception as e:
            print(f'Error instantiating template: {e}')

    def create_workflow_template(self):
        """! Create the workflow template, along with all relevant stages
        """
        raise NotImplementedError



#!/usr/bin/env python
# coding: utf-8

"""! @brief ZIDS_Predictable base class."""

from typing import Any
from dsframework.base.common import ZIDS_Object


class ZIDS_Predictable(ZIDS_Object):
    """! ZIDS_Predictable base class for a single predictable object."""

    def __init__(self) -> None:
        """! Initializer for ZIDS_Predictable
        Initializes local class variables, those ZIDS_Predictable variables will be transferred from component to
        component in the pipeline as an input and output.
        """
        ##
        # @hidecallgraph @hidecallergraph
        super().__init__()
        self.name:str = ''
        self.input:Any = None
        self.target:Any = None    
        self.pred: Any = None
        self.prob:float = -1.0
        self.forced_pred:Any = None
        self.forced_reason:str = 'Did Not Forced'



from locust import HttpUser, task
import pandas as pd
import numpy as np

from server.token_generator import tokenGenerator

##
# @file
# @brief Stress testing server using locust package.

t = tokenGenerator()
jwtToken = t.generateToken('STG_Conf')
global_headers = {
    'x-token': 'fake-super-secret-token',
    'Authorization': 'Bearer ' + jwtToken
}


class payloadUser(HttpUser):
    """! payloadUser class implements HttpUser base class. It is used to stress test a server, by sending repeated
    requests to the server, using locust package.

    Running a test:

    -# Install package, using:
        - pip install locust

    -# Create a sample dataset that reflects real data variety.
        - Save it in sample.csv file.
        - For example:
            @code {.py}
            titles_history,pred
            "[{"title":"Software Engineer","order":1,"desc":"","company":"","company_id":-1,"company_size":0},
            {"title":"Senior Software Engineer","order":2,"desc":"","company":"","company_id":-1,"company_size":0}]",0
            "[{"title":"verantwoordelijk persoon A. I.","order":1,"desc":"","company":"","company_id":-1,
            "company_size":0},{"title":"GDP responsible person","order":2,"desc":"","company":"","company_id":-1,
            "company_size":0}]",0
            @endcode
    -# Make a note of the expected load on the server - for example 2 requests per second.

    -# Run locust,
        - Set the following parameters:
            - --host - Host to load test in the following format: http://example.com
            - --users - Number of concurrent Locust users, we mostly test on 50, 100 and 200 users.
            - --spawn-rate - The rate per second in which users are spawned
        - locust -f locustfile.py --host={your-host} --users={number-of-users} --spawn-rate={your-spawn-rate}

    -# Access locust web-interface (gets created when locust runs), for example:
        - http://0.0.0.0:8089
        - Locus will display our selected parameters (modify if required).
    -# Start process:
        - Click 'Start Swarming' button.
    -# Statistics will show on screen:
        - Monitor the RPS (Results per second).
        - Make sure that the server can handle the expected load (i.e ~2 requests per second)
        - Take note of the maximum RPS that the server can handle before failures start to build up.
        If they do not match the expected results, consider modifying machine resources.
    -# Stop the test.
    -# Failures tab:
        - List of errors.
    -# Charts tab:
        - Shows the course of the test.
        - Download a pdf report and attached to release notes or PR notes.
    -# Things to keep in mind:
        - PC running the test - can affect the results.
        - Test for every new PR to see if results compromised.

    """

    def __init__(self, *args, **kwargs):
        """! payloadUser Initializer.

        Loads sample dataset from sample.csv file. """

        ##
        # @hidecallgraph @hidecallergraph
        super().__init__(*args, **kwargs)
        ## Sample dataset
        self.df = pd.read_csv("sample.csv", sep='\t')
        self.df.replace(np.nan, '', regex=True)

    def get_random_row(self):
        """Returns a random row from the sample dataset"""
        s = self.df.sample().to_dict('list')
        d = dict((k.lower(), v[0]) for k, v in s.items())
        return d

    @task
    def get_diagnosis_with_valid_payload(self):
        """The main method for stress testing, wrapped with locust @task decorator,
        it runs in a loop handled by locust."""
        headers = {"Content-Type": "application/json; charset=UTF-8", **global_headers}
        payload = self.get_random_row()
        self.client.post("/predict", json=payload, headers=headers)



import requests
import json
import sys
from server.token_generator import tokenGenerator

env = sys.argv[1]
model_name = sys.argv[2]
model_id = sys.argv[3]


t = tokenGenerator()
jwtToken = t.generateToken(env.upper() + '_Conf')

data = {
	"model_type_id": model_id
}
url = "https://dsp.zoominfo.com/api/v1/truth-datasets/getDatasetOptions"
if env == 'stg':
	url = "https://dsp-stg.zoominfo.com/api/v1/truth-datasets/getDatasetOptions"
headers = {
	'x-token': 'fake-super-secret-token',
	'Authorization': 'Bearer ' + jwtToken
}

response = requests.post(url, json=data, headers=headers)
try:
	result = json.loads(response.content)
	datasets = result['data']['datasets']
	dataset_options = [{'dataset_name': dataset['dataset_name'], 'id': dataset['id']} for dataset in datasets]
	if len(dataset_options) == 0:
		print(f'no dataset options for model named {model_name} in dsp')
	else:
		print(dataset_options)
except Exception as e:
	print('error', e)



"""! @brief Run pipeline stage, which will run the project's pipeline on the retrieved data."""
import sys
import pandas as pd
from typing import Any
from json import loads as json_loads
from pyspark.sql.types import *
from typing import Iterator

from dsframework.base.batch.stage_base import ZIDS_Stage


##
# @file
# @brief Stage main class, implements ZIDS_Stage base class.
class RunPipelineStage(ZIDS_Stage):
    """! Stage class

    Implement a stage that will later be converted to an executable job in a specific workflow.
    """

    def __init__(self, stage_config):
        """! The Stage class (generatedStageName) initializer.
        Base class will load basic configuration parameters, additional fields should be added here

            Args:
                stage_config : Configuration dictionary, loaded from configuration file.
        """

        ##
        # @hidecallgraph @hidecallergraph
        super().__init__(stage_config)
        self.udf_schema = StructType([StructField("msg", StringType()),
                                     StructField("output", StringType()),
                                     StructField("time_took", FloatType()),
                                     StructField("error", StringType()),
                                     StructField("url", StringType()),
                                     StructField("helper", StringType())])

    def get_stage_name(self):
        """! Get the stage name

            Returns:
                A string, containing the stage's name
        """
        return self.__class__.__name__

    def map_in_pandas(partition: Iterator[pd.DataFrame]):
        def test():
            from pipeline.pipeline import generatedProjectNamePipeline
            p = generatedProjectNamePipeline()
            output = p.execute(text='yuval')
            return output

        # def pandas_predict_trex_func(row):
        # print(f"{row=}")
        # print(f"{row.loc[0,'url']=}")
        # url_in = row.loc[0,'url']
        # out = {"msg":"sdfsdf","output":"outtt","time_took":4.4, "error": "errr", "url":url_in, "helper":"hellppme"   }
        # return out
        for pdf in partition:
            output = test()
            print('output', output)
        # print(f"{pdf=}")
        # print(f"{type(pdf)=}")
        # res = pandas_predict_trex_func(pdf)
        # res_pd  = pd.DataFrame([res])
        # yield res_pd

    def main(self, **kwargs: Any):
        """! Executes the main functionality of the stage.

            Args:
                **kwargs : Whatever is needed for the stage to run properly.
        """
        spark_session = self.load_spark()
        df = spark_session.createDataFrame([(1, 'http://aaa', "")], ("id", "url", "res"))
        out_df = df.mapInPandas(func=self.map_in_pandas, schema=self.udf_schema)
        print(out_df.to_string())


if __name__ == "__main__":
    """! Executes the stage by instantiating it and calling the main function.
    Set up argument condition according to the usage of the written stage

        Args:
            System argument 1 - Configuration file
            System argument 2 - Start date
    """
    if sys.argv and len(sys.argv) > 1:
        config = json_loads(sys.argv[1])
        stage = RunPipelineStage(config)
        if len(sys.argv) > 2:
            start_date = sys.argv[2]
            stage.update_start_date(start_date)
        stage.main()
    else:
        print(f"Stage configuration not provided, Can't run stage")





from ast import literal_eval
from flask import Flask, redirect, url_for, request, jsonify
from flask_cors import CORS, cross_origin
from dsframework.testable.dataset_base import DatasetSample
from generatedMainDirectory.tester import generatedProjectNameTester
import pandas as pd
import json

class generatedClass():

    def __init__(self):
        self.tester = generatedProjectNameTester()
        self.tester.init_pipeline()
        self.error_df = pd.DataFrame(columns=['signature', 'target'])
        self.report_df = None
        # self.report_col_names = self.tester.get_report_df_cols()
        # self.report_df = pd.DataFrame(columns=self.report_col_names)


    @staticmethod
    def run(debug=False):
        app.run(debug=debug)

    def handle_request(self, action, params,  truth_dataset_id=-1, model_type_id=-1, raw_id=-1):
        if action == 'parse':
            return self.parse_single(params)
        elif action == 'test':
            return self.test_mini_batch(params,  truth_dataset_id, model_type_id, raw_id)
        return {'action':action, 'params':params}

    def parse_single(self, params):
        text   = '' if 'text' not in params else params['text']
        uid    = '1'
        parsed = self.tester.parse(text)
        pred   = parsed['signatures'][0]
        pred['uid'] = uid
        return pred

    def test_mini_batch(self, params, truth_dataset_id=-1, model_type_id=-1, raw_id=-1):
        data = {} if 'data' not in params else params['data']
        data = literal_eval(data) if type(data) == str else data
        _id  = params['id'] if 'id' in params else -1
        signature_text = data['signature']
        target = data['segments']
        parsed = self.tester.parse(signature_text)
        pred   = parsed['signatures'][0]['segments']
        sample = DatasetSample(x=signature_text, y=target, pred=pred)
        results = self.tester.evaluate(sample)
        if self.report_df is None:
            self.report_df = pd.DataFrame(columns=self.tester.report_col_names)
        self.report_df.loc[len(self.report_df.index)] = results
        resp = {}
        try:
            for idx, col_name in enumerate(self.tester.report_col_names):
                if col_name in ['pred', 'target']:
                    results[idx] = json.dumps(results[idx])
                resp[col_name] = results[idx]
        except Exception as e:
            print('error', str(e),  '\n\n\n',signature_text, '\n\n\n', target, '\n\n\n', results , '\n\n\n')
            self.error_df.loc[len(self.error_df.index)] = [signature_text, target]
            self.error_df.to_csv('error_log.csv', index=False)

        resp['truth_id'] = _id
        resp['truth_dataset_id'] = truth_dataset_id
        resp['model_type_id'] = model_type_id
        resp['raw_id'] = raw_id
        if len(self.report_df.index)%100==0:
            self.report_df.to_csv('server_report.csv', index=False)
        return resp


app  = Flask(__name__)
cors = CORS(app)
app.config['CORS_HEADERS'] = 'Content-Type'
web_handler = generatedClass()

@app.route('/parse', methods = ['POST'])
@cross_origin()
def parse():
    if request.is_json:
        req    = request.json
        action = 'parse'
        text   = '' if 'text' not in req else req['text']
        params = {'text': text}
        resp   = [web_handler.handle_request(action, params)]
        return jsonify(resp)


@app.route('/test', methods = ['POST'])
@cross_origin()
def test():
    if request.is_json:
        req    = request.json
        action = 'test'
        truth_dataset_id = req['truth_dataset_id'] if 'truth_dataset_id' in req else -1
        model_type_id = req['model_type_id'] if 'model_type_id' in req else -1
        raw_id = req['raw_id'] if 'raw_id' in req else -1
        rows = req['rows'] if 'rows' in req else []
        resp = [web_handler.handle_request(action, params, model_type_id,truth_dataset_id,raw_id) for params in rows]
        return jsonify(resp)

if __name__ == '__main__':
    web_handler.run()





from dsframework.base.tester.evaluator_logic import ZIDS_Evaluator


class generatedClass(ZIDS_Evaluator):
    def __init__(self):
        """generatedClass initializer."""

        ##
        # @hidecallgraph @hidecallergraph
        super().__init__()

    def get_confusion_matrix(self, true_label, prediction):
        """
        Model evaluation function, to be implemented per project in evaluator.py file
        """
        raise NotImplementedError



import dsframework.base.pipeline.pipeline
import dsframework.base.pipeline.artifacts
import dsframework.base.pipeline.forcers
import dsframework.base.pipeline.postprocessor
import dsframework.base.pipeline.predictables
import dsframework.base.pipeline.predictors
import dsframework.base.pipeline.preprocessor



import sys
from typing import List

import pandas as pd
import os
import json
from random import randint

from pipeline.pipeline import generatedProjectNamePipeline
from pipeline.schema.inputs import generatedProjectNameInputs
from pipeline.schema.outputs import generatedProjectNameOutputs

from tester.test_schema.test_input import TestInputRequest, TestInputRow
from tester.test_schema.test_output import TestOutputResponse
from dsframework.base.tester.tester_logic import ZIDS_Tester
from dsframework.base.common.functions import flatten, split_dataframe, get_date_time
from tester.evaluator import generatedProjectNameEvaluator


##
# @file
# @brief generatedClass implementation of ZIDS_Tester base class.

class generatedClass(ZIDS_Tester):
    """Main purpose of this class is to support the 'local testing' feature in dsp, it allows to run the model
    on a selected dataset from dsp. It implements ZIDS_Tester.
    """
    def __init__(self):
        ZIDS_Tester.__init__(self)
        ## Not used
        self.ready = False
        ## Instance of the pipeline
        self.pipeline = generatedProjectNamePipeline()
        ## Instance of the Evaluator class
        self.evaluator = generatedProjectNameEvaluator()
        ## Fill this according to the field received from the model, and remove this comment
        self.pred_name = ""
        ## Fill this according to the field received from DSP, and remove this comment
        self.target_name = ""

    def test_batch(self, test_request: TestInputRequest) -> TestOutputResponse:
        """! This is the main function of the tester (Local testing feature in dsp), most of the functions
        are implemented in the base class (ZIDS_Tester), but wherever needed, a NotImplementedError will be raised.
        Feel free to mingle with the other functions as well. This is a humble suggestion.


        Args:
            test_request: A request for testing a batch, as received from the DSP (Data Science Portal)

        Returns:
            model_evaluation_results: A request formatted to fit to the DSP needs
        """

        # parse request rows as TestRow
        data_rows: List[TestInputRow] = test_request.rows

        if data_rows:
            # Convert the input to a dataframe input
            df_data_rows = self.convert_input_to_dataframe(data_rows)

            # Run the model on the data rows - implement create_model_request
            df_model_predictions = self.run_model_on_data_rows(df_data_rows)

            # Evaluate the results based on ground truth and predictions
            model_evaluation_results = self.evaluator.evaluate_model_predictions(df_model_predictions)

            # Format the evaluated output by changing the naming convention to fit with DSP
            model_evaluation_results = self.format_evaluation_output(model_evaluation_results)

        else:
            output = []
        # format response rows
        self.format_response_row_for_dsp_upload(model_evaluation_results, test_request)

        return model_evaluation_results

    def create_model_request(self, batch_input_rows) -> generatedProjectNameInputs:
        """! Structure the data as a proper request, and create a dictionary in order to enable serialization.

        Args:
            batch_input_rows: A batch of input information, intended to be structured into a model request
        Returns:
            model_request: The data, structured in a format that the model can work with
        """

        ##
        # This is an example code from the scoops project, please fit with your specific project fields:
        # @code{.py}
        # request_scoops = [ScoopCandidate(html_content=rec['text'],
        #                   source=rec['source'],
        #                   queue_name=rec['queue_name'],
        #                   title=rec['title']) for rec in batch_input_rows]
        # model_request = ScoopsClassificationRequest(pages=request_scoops)
        # @endcode

        raise NotImplementedError
        model_request = generatedProjectNameInputs()
        return model_request

    @staticmethod
    def create_mock_response(input_request: TestInputRequest) -> List[TestOutputResponse]:
        """! This function creates a mock output for the DSP to display.
        When starting a new project, this function can be used to test connection with the DSP.
        It of course has to be modified in order to reflect the correct API fields


        Args:
            input_request: Request as received from the DSP
        Returns:
            List[TestOutputResponse]: A list of responses, to be sent back to the DSP
        """
        input_request = input_request.dict()
        # print('DSP Request: ', input_request)

        response = []
        rows = input_request['rows']
        truth_dataset_id = input_request['truth_dataset_id']
        model_type_id = input_request['model_type_id']
        for row in rows:
            data = json.loads(row['data'])
            id = row['id']
            raw_id = row['raw_id'] or -1
            segments = data['segments'] if 'segments' in data else data
            pred = data['segments'] if 'segments' in data else data
            text = ''
            # Please feel free to adjust the mock response. This response fits the sigparser project
            if 'signature' in data:
                 text = data['signature']
            elif 'email' in data:
                 text = data['email']
            elif 'text' in data:
                text = data['text']
            elif 'name' in data:
                text = data['name']

            # Generate different confusion matrix values per row
            cm_list = [0, 0, 0, 0]
            cm_index = randint(0, 3)
            cm_list[cm_index] = 1

            response.append(
                TestOutputResponse(
                    **{
                        'truth_id': id,
                        'truth_dataset_id': truth_dataset_id,
                        'model_type_id': model_type_id,
                        'raw_id': raw_id,
                        'pred': json.dumps(pred),
                        'target': json.dumps(segments),
                        'text': text,
                        'name_first_fn': cm_list[0],
                        'name_first_fp': cm_list[1],
                        'name_first_pred': "Marty",
                        'name_first_prob': -1,
                        'name_first_target': "Marty",
                        'name_first_tn': cm_list[2],
                        'name_first_tp': cm_list[3],
                    }
                )
            )
        # call tester here

        return response

    def test_from_file(self, csv_file_path, batch_size):
        """! Run an evaluation from a csv file input.

        Args:
            csv_file_path: csv file containing both ground truth and raw input for the model.
            batch_size: Size of batch to use in order to feed the model with controllable chunks of data
        """
        ## Define prediction column name
        predictions_col_name = self.pred_name
        ## Define ground truth column name
        ground_truth_col_name = self.target_name

        # Define csv output
        time_str = get_date_time(False, True)
        csv_output_name = "model_eval_results_" + time_str + ".csv"
        csv_output_path = os.path.join(os.getcwd(), csv_output_name)
        first_chunk = True

        # Load CSV. predictions and ground truth columns are converted from string to list/dict
        df = pd.read_csv(csv_file_path,
                         converters={predictions_col_name: json.loads,
                                     ground_truth_col_name: json.loads})

        # Chunk according to batch size
        batches = split_dataframe(df, int(batch_size))

        # iterate over batches and send the data for evaluation
        for batch in batches:

            # Reset index of batch in order to align evaluation
            batch.reset_index(inplace=True)

            # Feed the model chunk by chunk
            output_rows = self.run_model_on_data_rows(batch)

            # Evaluate the results based on ground truth and predictions
            df_model_evaluation_results = self.evaluator.evaluate_model_predictions(output_rows)

            # Flatten confusion matrix and sort the dataframe
            conf_mat_dict = df_model_evaluation_results['confusion_matrix'].to_dict()
            confusion_matrix_flattened = flatten(conf_mat_dict)
            conf_pd = pd.DataFrame(confusion_matrix_flattened, index=[0])
            df_model_evaluation_results = df_model_evaluation_results.join(conf_pd)
            df_model_evaluation_results = df_model_evaluation_results.drop('confusion_matrix', 1)

            # Write output to csv, append the chunks
            if first_chunk:
                df_model_evaluation_results.to_csv(csv_output_path, mode='a')
                first_chunk = False
            else:
                df_model_evaluation_results.to_csv(csv_output_path, mode='a', header=False)


# This will call the tester with a given csv file
if __name__ == '__main__':

    tester = generatedClass()
    args = sys.argv[1:]
    tester.test_from_file(args[0], args[1])
    print("Done evaluating, please find the results at: model_eval_results.csv")



"""!
@mainpage Documentation

@section general_main General:
The dsframework was developed to unify and simplify our work by creating a structured work methodology. It allows you
to generate your own project out of a template which includes a default pipeline with the most common steps for
a data scientist to use.

@section benefits Benefits:
- Connection to ds-portal.
- Ready testing environment: local testing and cloud evaluation.
- Generate a pipeline with predefined four steps (that can be extended), it gives the flexibility for
a data scientist developer to load data, pass it through preprocessing procedures, a model, forcer and
run post process procedures.

@section notes_main Steps
The main default four steps included are:
- Preprocess
- Predictor
- Forcer
- Postprocessor

@section usg_main Start here:
For more information about:
 1. [Installing the ds-framework and generating a new project](QUICKSTART1.md) (@subpage quick_start1).
 2. [Setting up the project](QUICKSTART2.md) (@subpage quick_start2).
 3. [Test your project](QUICKSTART3.md) (@subpage quick_start3).
 4. [Additional cli commands, server endpoints and ignore files](MOREINFO.md) (@subpage more_info).


**Version** 1.0

Copyright (c) 2022 ZoomInfo.  All rights reserved.
"""


import pandas as pd
import json
from typing import Dict, List
from dsframework.base.common.functions import flatten

##
# @file
# @brief ZIDS_Tester base class for general_tester.py

class ZIDS_Tester:
    """ZIDS_Tester class base class for general_tester.py, contains methods to format data to Inputs, Outputs and dsp,
        requirements.
    """
    def __init__(self):
        """! ZIDS_Tester initializer.

        Initializes local variables.
        """
        self.truth_dataset_id = -1
        self.model_type_id = -1
        self.data_keys = []
        ## Prediction field name
        self.pred_name = ""
        ## Ground truth fields name - implemented in derived class.
        self.target_name = ""

    def save_meta_data(self, input_req):
        """! Save input request metadata, truth_dataset_id and model_type_id.

        Args:
            input_req: An object of type TestInputRequest
        """

        ## Truth dataset ID
        self.truth_dataset_id = input_req.truth_dataset_id
        ## Model type ID
        self.model_type_id = input_req.model_type_id

    def test_batch(self, test_request):
        """Implemented in derived class generatedClassGeneralTester in file general_tester.py"""
        raise NotImplementedError

    def convert_input_to_dataframe(self, data_rows):
        """! Convert from a list of TestInputRow to a dataframe.
        - Used in general_tester.py test_batch()
        - Also apply Model-Specific conversion.


        Args:
            data_rows: A list of objects, List[TestInputRow]
        Returns:
            Pandas dataframe that contains all the relevant information
        """
        df = pd.DataFrame.from_records((data_row.dict() for data_row in data_rows))
        df = df.apply(self.adapt_to_dsp_format, axis=1)
        return df

    def create_model_request(self, batch_input_rows):
        """Implemented in derived class generatedClassGeneralTester in file general_tester.py"""
        raise NotImplementedError

    def run_model_on_data_rows(self, row_dataframe: pd.DataFrame) -> pd.DataFrame:
        """! Runs model on all data rows. Create a format, the model can work with and runs the pipeline.

        Args:
            row_dataframe: Input dataframe, containing all data rows.
        Returns:
            Extended dataframe which contains the output of the model
        """
        batch_of_pages = row_dataframe.to_dict(orient='records')

        model_request = self.create_model_request(batch_of_pages)

        # Call the model to get predictions
        predicted_records = self.pipeline.execute(model_request)

        if predicted_records:
            df_predicted = self.create_predicted_records_dataframe(predicted_records)
            row_dataframe = row_dataframe.join(df_predicted)
        else:
            # we did not get any records, create empty records
            self.create_empty_predictions(row_dataframe)

        return row_dataframe

    def adapt_to_dsp_format(self, df_row):
        """! Adapt for DSP format compatibility
        'data' is provided as string since the DSP saves it this way.
        Please feel free to modify due to your project needs.


        Args:
            df_row: A single row from the processed data frame
        Returns:
            df_row: The flattened data frame, with all the fields previously compressed in the 'data' field
        """

        ##  Extract keys for later use
        self.data_keys = json.loads(df_row['data']).keys()

        # Restructure as a data frame line
        for k, v in json.loads(df_row['data']).items():
            df_row[k] = v
        return df_row

    def create_predicted_records_dataframe(self, predicted_records):
        """! Create a dataframe formatted as input for evaluation

        Args:
            predicted_records: Results from pipeline
        Returns:
            df_predicted: DataFrame
        """

        df_predicted = pd.DataFrame(predicted_records)
        df_predicted = df_predicted.rename(columns={self.pred_name: 'pred_model'})
        return df_predicted

    @staticmethod
    def create_empty_predictions(row_dataframe):
        """! Create empty predictions in case the model returns an error, and we wish to keep evaluating.
        Not implemented yet.


        Args:
            row_dataframe: Input dataframe, containing all data rows.
        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def format_evaluation_output(self, row_dataframe) -> List[Dict]:
        """! Format the results with the expected DSP data format.

        Args:
            row_dataframe: Results from evaluation
        Returns:
            List[Dict]: In dsp format
        """

        row_dataframe = row_dataframe.rename(columns={'pred': 'target', 'pred_model': 'pred'})

        output = row_dataframe.to_dict(orient='records')
        for rec in output:
            pred_dict = {}
            target_dict = {}
            for row_key in self.data_keys:
                pred_dict[row_key] = rec[row_key]
                target_dict[row_key] = rec[row_key]
            pred_dict['pred'] = rec['pred']
            target_dict['pred'] = rec['target']

            for row_key in self.data_keys:
                del rec[row_key]

            rec['target'] = json.dumps(target_dict)
            rec['pred'] = json.dumps(pred_dict)

        return output

    @staticmethod
    def format_response_row_for_dsp_upload(output, test_request):
        """! Format response rows (including results) to follow DSP requirements.

        Args:
            output: model_evaluation_results
            test_request: Include ids required for response - truth_dataset_id, model_type_id -
        Returns:
            Doesn't return, make changes by reference to input argument 'output'.
        """
        row_extra_data = {
            "truth_dataset_id": test_request.truth_dataset_id,
            "model_type_id": test_request.model_type_id
        }
        for row in output:
            row.update(row_extra_data)
            row['truth_id'] = row.get('id', -1)

            if 'id' in row:
                del row['id']

            confusion_matrix_flattened = flatten(row['confusion_matrix'])
            row.update(confusion_matrix_flattened)

            del row['confusion_matrix']

            if 'id' in row:
                del row['data']

    @staticmethod
    def get_f1_precision_recall(var_name, df):
        """! Calculate from detailed output report the f1, precision and recall scores per attribute.

        This function should be overridden if other form of calculation is needed

        Args:
            var_name: variable name to calculate the scores for.
            df: detailed results dataframe with per attribute confusion matrix as outputted by tester.
        Returns:
            f1, precision and recall scores.

        """
        tp = df[var_name + '_tp']
        fp = df[var_name + '_fp']
        tn = df[var_name + '_tn']
        fn = df[var_name + '_fn']
        recall = 0
        precision = 0
        f1 = 0
        if len(np.where(tp + fp + tn + fn != 1)[0]) > 0:
            print(f'{len(np.where(tp + fp + tn + fn != 1)[0])} examples have errors')
        if (tp.sum() + fn.sum()) != 0:
            recall = tp.sum() / (tp.sum() + fn.sum())
        if (tp.sum() + fp.sum()) != 0:
            precision = tp.sum() / (tp.sum() + fp.sum())
        if (precision + recall) != 0:
            f1 = 2 * ((precision * recall) / (precision + recall))
        return f1, precision, recall


from dsframework.base.common.component import ZIDS_Component
from dsframework.base.common import functions
from dsframework.base.common.object import ZIDS_Object
from dsframework.base.common.regex_handler import RegexHandler



"""! @brief Set Data stage, writing obtained data onto a database."""
import sys
from typing import Any
from json import loads as json_loads
from google.cloud import bigquery

from dsframework.base.batch.stage_base import ZIDS_Stage


##
# @file
# @brief Stage main class, implements ZIDS_Stage base class.
class SetDataStage(ZIDS_Stage):
    """! Stage class

    Implement a stage that will later be converted to an executable job in a specific workflow.
    """

    def __init__(self, stage_config):
        """! The Stage class (generatedStageName) initializer.
        Base class will load basic configuration parameters, additional fields should be added here

            Args:
                stage_config : Configuration dictionary, loaded from configuration file.
        """

        ##
        # @hidecallgraph @hidecallergraph
        super().__init__(stage_config)
        self.start_date = ""
        if 'user_email' in stage_config:
            self.user_email = stage_config['user_email']

        if self.unique_iteration_id:
            self.folder_path = self.project_name + '/' + self.user_email + \
                               '/unique_iteration_id_' + self.unique_iteration_id
        else:
            self.folder_path = self.project_name + '/' + self.user_email + '/main'

        self.bucket_path = f'gs://{self.bucket_name}/{self.folder_path}'

    def set_start_date(self, start_time):
        """! Setting start date, if available.
            Args:
                start_time - Stage start time
        """
        print(f'Setting start date: {start_time}')
        self.start_date = start_time

    def get_stage_name(self):
        """! Get the stage name

            Returns:
                A string, containing the stage's name
        """
        return self.__class__.__name__

    def main(self, **kwargs: Any):
        """! Executes the main functionality of the stage.

            Args:
                **kwargs : Whatever is needed for the stage to run properly.
        """
        ds_source = "zido-prd-svpc-4"
        date_str = "2021-10-12"
        inputParquetName = "stage_load_join_ner_data"
        outputParquetName = "stage_load_join_ner_data"

        spark = self.load_spark()

        page_list = spark.read.parquet(f"{self.bucket_path}/data/{inputParquetName}.parquet")
        page_list.printSchema()
        page_list.show(5)

        print("X write bigquery X")
        # Update to your GCS bucket
        bq_dataset = self.project_name.replace('-', '_')
        # bq_table = f'stage_crawler_list_{config.limit_table}_{str(config.sample).replace(".", "p")}'
        bq_table = f'stage_crawler_list_latest_batch'

        print(f"{bq_table=}")

        client = bigquery.Client()
        dataset_id = "{}.{}".format(self.project_id, bq_dataset)
        dataset_exist = False

        try:
            client.get_dataset(dataset_id)
            dataset_exist = True
        except:
            pass

        if not dataset_exist:
            dataset = bigquery.Dataset(dataset_id)
            dataset.location = self.region
            dataset = client.create_dataset(dataset, timeout=30)  # Make an API request.
            print("Created dataset {}.{}".format(client.project, dataset.dataset_id))

        page_list.write \
            .format("bigquery") \
            .option("table", "{}.{}".format(bq_dataset, bq_table)) \
            .option("temporaryGcsBucket", self.bucket_name) \
            .mode('overwrite') \
            .save()


if __name__ == "__main__":
    """! Executes the stage by instantiating it and calling the main function.
    Set up argument condition according to the usage of the written stage

        Args:
            System argument 1 - Configuration file
            System argument 2 - Start date
    """
    if sys.argv and len(sys.argv) > 1:
        config = json_loads(sys.argv[1])
        stage = SetDataStage(config)
        if len(sys.argv) > 2:
            start_date = sys.argv[2]
            stage.update_start_date(start_date)
        stage.main()
    else:
        print(f"Stage configuration not provided, Can't run stage")





import json
import pickle
from datetime import datetime

from dsframework.base.common.regex_handler import RegexHandler

regex = RegexHandler

##
# @file
# @brief This file has general functions that can simplify things when coding.



def flatten(d,sep="_"):
    """! Flatten dictionary recursively
       For example   : d = {"message": {"hello": 123456}}
        will return -> obj = {"message_hello": 123456}
      Args:
        d: the dictionary we want to flatten
      Returns:
        obj : break all dict in dict and return one flatten dict
    """
    import collections

    obj = collections.OrderedDict()

    def recurse(t, parent_key=""):

        if isinstance(t, dict):
            for k, v in t.items():
                recurse(v, parent_key + sep + k if parent_key else k)
        else:
            obj[parent_key] = t
    recurse(d)
    return obj


def get_from_dict(key, container:dict, default='', delete=False):
    """! Get value from the specific dictionary[key].
         Also, can delete item
        Args:
          key : dictionary key.
          container: the dictionary.
          default : empty string ('').
          delete(=False) : if TRUE delete container[key]
        Returns:
           val : from container[key] .
    """
    if not container or type(container) is not dict:
        return default
    if key in container:
        val = container[key]
        if delete:
            del container[key]
        return val
    return default


def get_date_time(short=False, no_space=False):
    """! Get date & time in required form
         default form example: 28/02/2022 11:59:59
      Args:
          short(default=False): only get date (example: 28_02_2022).
          no_space(default=False): No space (example: 20220328-0:00:59)
      Returns:
          string : which includes the time&date with the required form.
      """
    now = datetime.now()
    form = '%d/%m/%Y %H:%M:%S'
    if short:
        form = '%d_%m_%Y'
    elif no_space:
        form = "%Y%m%d-%H%M%S"
    dt_string = now.strftime(form)
    return dt_string


def load_pickle(path):
    """! Loads pickle file
     Args:
         path: Absolute path to data file.
     Returns:
         pkl : pickle file.
     """

    pkl = None
    with open(path, 'rb') as fid:
        pkl = pickle.load(fid)
    return pkl


def load_json(path):
    """! Loads json file
     Args:
         path: Absolute path to the file.
    Returns:
         jsn : json file.
    """
    jsn = None
    with open(path) as fid:
        jsn =  json.load(fid)
    return jsn 

def load_file_to_dict(in_file, sep=',', key_type=str, value_type=str):
    """! Loads file and parse him to a dictionary with (key , value )
     Args:
         in_file : the file we want to parse.
         sep (=',') : character separation
         key_type : The key type.
         value_type : The value type.
     Returns:
         dictionary : The dictionary created from the file.
     """
    dictionary = {}
    with open(in_file) as f:
        for line in f:
            (key, value) = line.rstrip('\n').split(sep)
            dictionary[key_type(key)] = value_type(value)
    return dictionary

def save_pickle(obj, path):
    """! Saves as a pickle
      Args:
          obj : the object we want to add to the file.
          path: Absolute path to the file.
    """
    with open(path, 'wb') as fid:
        pickle.dump(obj, fid)


def save_json(obj, path):
    """! Saves as a json
       Args:
           obj : the object we want to add to the file.
           path: Absolute path to the file.
     """
    with open(path, 'w') as f:
        json.dump(f, obj)


def remove_empty_leafs(d):
    """! Remove empty leafs from list or dictionary with recursion
       Args:
           d : list or dictionary.
       Returns:
           d : the same input without empty nodes.
    """
    if not isinstance(d, (dict, list)):
        return d
    if isinstance(d, list):
        return [v for v in (remove_empty_leafs(v) for v in d) if v]
    return {k: v for k, v in ((k, remove_empty_leafs(v)) for k, v in d.items()) if v}


def flatten_list(l):
    """! This function Flatten list and return it.
    @verbatim
    Args:
        l : List.
    Returns:
        list : flatten list
    @endverbatim
    """
    def get_el(el):
        if type(el) is not list:
            return [el]
        return el
    l2 = [get_el(el) for el in l]
    flatten = lambda l: [item for sublist in l for item in sublist]
    return flatten(l2)


def is_list_of_list(l):
    """! Is list of list bool function.
    @verbatim
    Args:
        l : List.
    Returns:
        bool: True if successful, False otherwise.
    @endverbatim
    """
    for el in l:
        if type(el) is list:
            return True
    return False


def ngram_splitter(line, ngram=3,all_grams=False):
    """! NGRAM splitter.
    @verbatim
    Args:
        line : Line .
        ngram (=3) :
        all_grams(=False) :
    Returns:
        string : splitted line (regarding the ngram number input)
    @endverbatim
    """
    tokens       = line.split(" ")
    loop_index   = 1
    result       = []
    ngram_window = []
    for token in tokens:
        ngram_window.append(token)
        if all_grams==False:
            if loop_index >= ngram:
                result.append(" ".join(ngram_window))
        else:
            for j in range(1,ngram+1):
                if loop_index >=j:
                    if len(result)<j:
                        result.append([" ".join(ngram_window[0:loop_index-1+j])])
                    else:
                        result[j-1].append(" ".join(ngram_window[min(loop_index-j,ngram-j):min(ngram,loop_index+j-1)]))
        if loop_index >= ngram:
            ngram_window = ngram_window[1:]
        loop_index +=1 
    return result


def remove_duplicates_lines(lines = []):
    """! Remove duplicates lines.
    @verbatim
    Args:
       lines : all the lines
    Returns:
       list : lines that are uniq from each other.
    @endverbatim
    """
    ls = lines
    lookup = set()  # a temporary lookup set
    ls = [x for x in ls if x not in lookup and lookup.add(x) is None]
    return ls 


def get_html_block_elements_list():
    """! This function returns the HTML block elements as a list.
        for example : ['<address', '<article', '<aside', '<blockquote',...]
    @verbatim
    Returns:
       list : HTML block elements.
    @endverbatim
     """
    return ['<address', '<article', '<aside', '<blockquote', '<canvas', '<dd', '<div',
            '<dl',
            '<dt', '<fieldset', '<figcaption', '<figure', '<footer', '<form', '<h1',
            '<h2', '<h3',
            '<h4', '<h5', '<h6', '<header', '<hr', '<li', '<main', '<nav',
            '<noscript', '<ol', '<p', '<pre', '<section', '<table', '<tfoot', '<ul',
            '<video']


def get_html_inline_elements_list():
    """! This function returns the HTML inline elements as a list.
        for example : ['<code', '<dfn', '<em',...]
    @verbatim
    Returns:
       list : HTML inline elements.
    @endverbatim
    """
    return ['<a', '<abbr', '<acronym', '<b', '<bdo', '<big', '<br', '<button', '<cite',
            '<code', '<dfn', '<em',
            '<i', '<img', '<input', '<kbd', '<label', '<map', '<object', '<output',
            '<q', '<samp', '<script',
            '<select', '<small', '<span', '<strong', '<sub', '<sup', '<textarea',
            '<time', '<tt', '<var']


def split_dataframe(df, batch_size=3):
    """
    Helper function that can chunk a dataframe according to a given chunk size
    :param df: Dataframe to divide
    :param batch_size: Batch size
    :return: A list of chunks, each the size of the desired batch size
    """
    chunks = list()
    num_chunks = len(df) // batch_size + 1
    for i in range(num_chunks):
        chunks.append(df[i * batch_size:(i + 1) * batch_size])
    return chunks




#!/usr/bin/env python
# coding: utf-8
"""! @brief ZIDS_Object base class."""


class ZIDS_Object():
    """! ZIDS_Object base class for ZIDS_Predictable.
    by default it is not implemented.
    Can be used for declaring variables that needs to shared between predictable objects."""
    pass

    


import os
import subprocess
from textwrap import dedent

from dsframework.base.batch.dag_base import ZIDS_Dag


class Dag(ZIDS_Dag):
    def __init__(self, dag_config_params):
        """! Loads config file and initializes parameters.

            Args:
                dag_config_params: Configuration dictionary, loaded from file
        """

        # dag params from airflow 1 - https://airflow.apache.org/docs/apache-airflow/1.10.3/macros.html
        # dag params from airflow 2 - https://airflow.apache.org/docs/apache-airflow/stable/templates-ref.html
        super().__init__(dag_config_params)

        if self.unique_template_id:
            self.orig_project_name = self.orig_project_name + '_' + self.unique_template_id

        # Create DAG python code file
        self.file_name = self.orig_project_name + "_dag.py"
        text_file = open(self.file_name, "w")
        n = text_file.write(dedent(self.get_basic_dag_code()))
        text_file.close()

    def create_dag(self):
        """! Dag main function.
        This function is the "entrypoint" for the dag creation.
        """

        source = self.file_name

        command = '''gcloud composer environments storage dags import \
            --environment "''' + self.environment + '''" \
            --location "''' + self.region + '''" \
            --source "''' + source + '''" \
            --destination "''' + self.orig_project_name + '''"
        '''
        # print(command)
        process = subprocess.call(command, shell=True)

        os.unlink(self.file_name)
        dag_path = self.bucket_name + '/dags/' + self.orig_project_name + '/' + source
        print(f'finish uploading dag to - {dag_path}')


if __name__ == "__main__":
    """! Triggers the creation of the workflow dag, which will instantiate a
    workflow template into an actual workflow, executing it.
    """

    dag = Dag()
    dag.create_dag()



from dsframework.base.common.component import ZIDS_Component

##
# @file
# @brief Forcer class, implements ZIDS_Component base class.
class generatedClass(ZIDS_Component):
    """! Forcer class, implements ZIDS_Component base class.

        Class was added with two main goals, to force results when:

        1. Certain data doesn't need to go through the model prediction, and we would like to force other results.
        2. Not satisfied by some results received from the model and would like to force other result.

        No examples yet.
    """
    def __init__(self, artifacts=None) -> None:
        """! generatedClass (Forcer) class initializer."""

        ##
        # @hidecallgraph @hidecallergraph
        super().__init__(artifacts)



#!/usr/bin/env python
# coding: utf-8

##
# @file
# @brief ZIDS_Postprocessor base class for Postprocess class.

from typing import List, Any
from dsframework.base.common.component import ZIDS_Component
from dsframework.base.pipeline.predictables.predictable import ZIDS_Predictable
from dsframework.base.pipeline.artifacts.shared_artifacts import ZIDS_SharedArtifacts

class ZIDS_Postprocessor(ZIDS_Component):
    """! ZIDS_Postprocessor the base class for Postprocess class."""

    def __init__(self, artifacts:ZIDS_SharedArtifacts=None) -> None:
        """! ZIDS_Postprocessor initializer

        Args:
            artifacts(ZIDS_SharedArtifacts): Shared artifacts instance.
        """

        ##
        # @hidecallgraph @hidecallergraph
        super().__init__(artifacts)

    def normalize_output(self, predictables:List[ZIDS_Predictable]) -> Any:
        """! ZIDS_Postprocessor.normalize_output base method.

        Not implemented, override in generatedProjectNamePostprocess class.

        This method should return final results output as a generatedProjectNameOutputs datatype object.

        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def execute(self,  predictables:List[Any]) -> Any:
        """! Executes postprocess step.

        Args:
            predictables:List[Any]: List of predictables received from the model and/or forcer.
        Returns:
            List[Outputs]: List of results after post-processing.
        """
        return self.normalize_output(super().execute(predictables))



#!/usr/bin/env python
# coding: utf-8

##
# @file
# @brief This file hold all regular expressions used in the project.
#        We declare the pattern and compile them for better performance.
#        We wrap each regex / group of regexes with functions as API
#        For example RegexHandler.url2.findall(text) -> find all urls in the text
#        in the background we use
#        RegexHandler.url2 = re.compile(url_pattern2)
#        url_pattern2 = r"(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'\".,<>?«»“”‘’]))"
#
#        Another example with emojis:
#        ----------------------------
#        we define:
#        emoji_pattern = u'([\U00002600-\U000027BF])|([\U0001f300-\U0001f64F])|([\U0001f680-\U0001f6FF])'
#        emoji = re.compile(emoji_pattern, flags=re.UNICODE)
#        and then we can use:
#        emojis = RegexHandler.emoji.findall(t)  // -> in order to find emojis
#        or
#        return RegexHandler.emoji.sub(r'', t)  // -> in order to remove all emojis from text

import datetime
import math
import re
import string
import sys
import unicodedata
from string import punctuation
from unidecode import unidecode
import tldextract

class RegexHandler:
    def __init__(self):
        pass

    url_pattern = r'''(?i)\b((?:[a-z][\w-]+:(?:/{1,3}|[a-z0-9%])|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'".,<>?«»“”‘’]))'''
    url_pattern2 = r"(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'\".,<>?«»“”‘’]))"
    ## For example 36 steps , 0.1ms  (https://www.example.com/)
    url_pattern3 = r"((http|https):\/\/)(www.)?[a-zA-Z0-9@:%._\\+~#?&\/\/=]{2,256}\.[a-z]{2,6}([-a-zA-Z0-9@:%._\\+~#?&\/\/=]*)"
    email_pattern = r'([\w\.-]+@[\w\.-]+\.\w+)'
    emoji_pattern = u'([\U00002600-\U000027BF])|([\U0001f300-\U0001f64F])|([\U0001f680-\U0001f6FF])'
    number_pattern = r'\d+'
    join_spaces_pattern = r' {2,}'
    join_nl_pattern = r'\n{2,}'
    html_pattern = r'<.*?>'
    time_pattern = r'^(([01]\d|2[0-3]):([0-5]\d)|24:00)$'
    has_time_pattern = r'(\d+):(\d+)'
    contact_prefixes_pattern = ['email', 'linkedin', 'twitter', 'mail']
    contact_prefixes_pattern = '(' + '|'.join(contact_prefixes_pattern) + ')[: ]+'
    connections_hyphen_pattern = r'(\d+)[\-]?(direct|office|main|fax|mob|cel)'
    tel_patterns = r'(toll\s?\-?free:|general:|general tel:|helpline:|line:|call:|main:|customer service:|general line:)\s*(\+?\d+)'
    direct_pattern = r'(direct call\:?)\s*(\+?\d+)'
    toll_free_pattern = r'toll[\s\-] free (service|tel)'
    no_number_pattern = r'(\s+)(no:|number:)(\+?\d+)'
    po_box1_pattern = r'\np\.?o\.?\s*?box\.?\s*?\:?\n(\d+)'
    po_box2_pattern = r'p\.?o\.?\s*?box\.?\s*?\:\s*(\d+)'
    follow_me_pattern = r'follow me on'
    since_pattern = r'since \d{4}'
    all_rights_pattern = r'ALL RIGHTS RESERVED'
    hours_prefix_pattern = r'(office|operating|branch|working) (hours|hrs).*(\n|$)'
    offices_pattern = r' (offices\:?\s+)'
    corporate_pattern = r'corporate (office\:?)'
    plus_phone_pattern1 = r'(\d{3,20})\/(\+\d+)'
    plus_phone_pattern2 = r'(\D+:)(\+?\d+)'
    tel_link_pattern = r'tel:([\s]?\d{3,20})'
    dot_list_pattern = r'\n(\d\.\s+)(\w+)'
    direct_line_dial_pattern = r'direct (line|dial)'
    main_line_pattern = r'(main line)|(HelpTel)'
    main_office_pattern = r'main office'
    email_prefix_pattern = r'[eE]-mail'
    skype_prefix_pattern = r'skype:'
    separator_pattern = r' by |│| • | · | I | l '
    url_junk_pattern = r'[0-9a-fA-F%/-]'
    # split_by_separator_pattern = r' *[\|~] *'
    semicolon_spaces_pattern = r'\s+(:|,)'
    digits_brackets_pattern = r'\(\s?(\+?\d+)\s?\)'
    rep_with_space_pattern = r'(\*|~|\|)'
    number_with_x_pattern = r'\bx(\d+)\b'
    concat_phones_pattern1 = r'(\d{6,20}/\d*)\s+(\d{8,20})'
    concat_phones_pattern2 = r'(\d{5,20})\s+(\d{3}-\d+\D*)'
    concat_phones_pattern3 = r'(\d{3,4}-\d{3,4})\s+(\d+-\d+)'
    concat_phones_pattern4 = r'(\d{6,20})\s+(\d{5,20})'
    ref_pattern = r'ref[\s]?\d+'

    ref = re.compile(ref_pattern)
    concat_phones1 = re.compile(concat_phones_pattern1)
    concat_phones2 = re.compile(concat_phones_pattern2)
    concat_phones3 = re.compile(concat_phones_pattern3)
    concat_phones4 = re.compile(concat_phones_pattern4)
    number_with_x = re.compile(number_with_x_pattern)
    rep_with_space = re.compile(rep_with_space_pattern)
    digits_brackets = re.compile(digits_brackets_pattern)
    semicolon_spaces = re.compile(semicolon_spaces_pattern)
    tel = re.compile(tel_patterns, flags=re.I)
    direct = re.compile(direct_pattern, flags=re.I)
    connections_hyphen = re.compile(connections_hyphen_pattern, flags=re.I)
    toll_free = re.compile(toll_free_pattern, flags=re.I)
    no_number = re.compile(no_number_pattern, flags=re.I)
    po_box1 = re.compile(po_box1_pattern, flags=re.I)
    po_box2 = re.compile(po_box2_pattern, flags=re.I)
    follow_me = re.compile(follow_me_pattern, flags=re.I)
    offices = re.compile(offices_pattern)
    since = re.compile(since_pattern, flags=re.I)
    all_rights = re.compile(all_rights_pattern, flags=re.I)
    hours_prefix = re.compile(hours_prefix_pattern, flags=re.I)
    corporate = re.compile(corporate_pattern, flags=re.I)
    plus_phone1 = re.compile(plus_phone_pattern1, flags=re.I)
    plus_phone2 = re.compile(plus_phone_pattern2, flags=re.I)
    tel_link = re.compile(tel_link_pattern)
    dot_list = re.compile(dot_list_pattern, flags=re.I)
    direct_line_dial = re.compile(direct_line_dial_pattern, flags=re.I)
    main_line = re.compile(main_line_pattern, flags=re.I)
    main_office = re.compile(main_office_pattern, flags=re.I)
    email_prefix = re.compile(email_prefix_pattern, flags=re.I)
    skype_prefix = re.compile(skype_prefix_pattern, flags=re.I)
    separator = re.compile(separator_pattern)
    url_junk = re.compile(url_junk_pattern)
    # split_by_separator = re.compile(split_by_separator_pattern)
    contact_prefixes = re.compile(contact_prefixes_pattern,flags=re.I)


    url = re.compile(url_pattern)
    url2 = re.compile(url_pattern2)
    url3 = re.compile(url_pattern3)
    email = re.compile(email_pattern)
    emoji = re.compile(emoji_pattern, flags=re.UNICODE)
    number = re.compile(number_pattern)
    merge_lines = re.compile(join_nl_pattern)
    merge_spaces = re.compile(join_spaces_pattern)
    html = re.compile(html_pattern)
    time = re.compile(time_pattern)
    has_time = re.compile(has_time_pattern)
    camelCase1 = re.compile(r'([A-Z][a-z]+)')
    camelCase2 = re.compile(r'([A-Z]+)')
    camelCase3 = re.compile(r'(.)([A-Z][a-z]+)')
    camelCase4 = re.compile(r'([a-z0-9])([A-Z])')
    camelCase5 = re.compile(r'([A-Z])_([A-Z]+[0-9]*[A-Z]+)')
    bad_decoding = re.compile(r'^(\s*=?[A-F][A-F0-9]=?\s*)+$')
    # square_brackets = re.compile(r'\[[^)^\]]*\]')
    brackets_cid = re.compile(r'\(cid:[^)]*\)')
    usd_amount = re.compile(r'\$\d{2,}\.?\d*')

    keywords = []

    word2num = {'zero': 0, 'one': 1, 'two': 2, 'three': 3, 'four': 4, 'five': 5,
                'six': 6, 'seven': 7, 'eight': 8, 'nine': 9, 'ten': 10,
                'eleven': 11, 'twelve': 12, 'thirteen': 13, 'fourteen': 14,
                'fifteen': 15, 'sixteen': 16, 'seventeen': 17, 'eighteen': 18, 'nineteen': 19,
                'twenty': 20, 'thirty': 30, 'forty': 40, 'fifty': 50, 'sixty': 60, 'seventy': 70, 'eighty': 80,
                'ninety': 90
                }

    months = ['january', 'jan', 'february', 'feb', 'march', 'mar', 'april', 'apr']
    months += ['may', 'june', 'jun', 'july', 'jul', 'august', 'aug', 'september', 'sep', 'sept']
    months += ['october', 'oct', 'november', 'nov', 'december', 'dec']

    days = ['monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday']
    days += ['mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun']

    # build a table mapping all non-printable characters to None
    NOPRINT_TRANS_TABLE = {
        i: None for i in range(0, sys.maxunicode + 1) if not chr(i).isprintable()
    }

    @staticmethod
    def remove_ref(s):
        """
        handles cases of digits followed by one of direct|office|main|fax|mob|cel (with/without "-" between)
        it will remove the "-" if exists and separate the number and the word with a space
        @param s: string
        @return: string
        """
        new_text = []
        for l in s.split("\n"):
            if len(RegexHandler.ref.findall(l.lower())) > 0:
                continue
            else:
                new_text.append(l)
        return "\n".join(new_text)

    @staticmethod
    def sub_connections_hyphen(s):
        """
        handles cases of digits followed by one of direct|office|main|fax|mob|cel (with/without "-" between)
        it will remove the "-" if exists and separate the number and the word with a space
        @param s: string
        @return: string
        """
        return RegexHandler.connections_hyphen.sub(r'\1 \2', s)

    @staticmethod
    def sub_separators(s):
        """
        replaces different types of separators to "|"
        @param s: string
        @return: string
        """
        return RegexHandler.separator.sub(' | ', s)

    @staticmethod
    def sub_main_office(s):
        """
        Part of the text_formatter. replaces main office to "Office"
        @param s: string
        @return: string
        """
        return RegexHandler.main_office.sub('Office ', s)

    @staticmethod
    def sub_skype_email_prefix(s):
        """
         Part of the text_formatter. Replaces email/E-mail/e-mail to Email, and also all types of skype (SKYPE) to Skype
        @param s: string
        @return: string
        """
        s = RegexHandler.email_prefix.sub('Email ', s)
        return RegexHandler.skype_prefix.sub('Skype: ', s)

    @staticmethod
    def sub_direct_line_dial(s):
        """
        Part of text_formatter. Replaces Direct line/dial to "Direct "
        @param s: string
        @return: string
        """
        return RegexHandler.direct_line_dial.sub('Direct ', s)

    # noinspection PyMethodParameters
    def sub_main_line(s):
        """
        Part of text_formatter. Replaces main line to "Tel "
        @param s: string
        @return: string
        """
        return RegexHandler.main_line.sub('Tel ', str(s))

    @staticmethod
    def sub_tel_link(s):
        """
        Part of text_formatter. handles cases of <a tel:123456 />
        @param s: string
        @return: string
        """
        return RegexHandler.tel_link.sub(r' \1', s)

    @staticmethod
    def sub_dot_list(s):
        """
        Part of text_formatter. handles cases such:
        "1. text" -> " text"
        @param s: string
        @return: string
        """
        return RegexHandler.dot_list.sub(r'\n \2', s)

    @staticmethod
    def sub_plus_phone(s):
        """
        Part of text_formatter. handles cases such:
        "+12345678" -> "12345678"
        @param s: string
        @return: string
        """
        s = RegexHandler.plus_phone1.sub(r'\1 \2', s)
        return RegexHandler.plus_phone2.sub(r'\1 \2', s)

    @staticmethod
    def sub_offices_corporate(s):
        s = RegexHandler.offices.sub(r'\n \1', s)
        return RegexHandler.corporate.sub(r' \1', s)

    @staticmethod
    def sub_hours_prefix(s):
        """
        Part of text_formatter. Replaces prefix before hours (office|working|etc..) to ''
        @param s: string
        @return: string
        """
        return RegexHandler.hours_prefix.sub('', s)

    @staticmethod
    def sub_non_sig_text(s):
        """
        Part of text_formatter. Removed non relevant text (follow me on| all rights reserved | etc..)
        @param s: string
        @return: string
        """
        s = RegexHandler.follow_me.sub('', s)
        s = RegexHandler.since.sub('', s)
        return RegexHandler.all_rights.sub('', s)

    @staticmethod
    def sub_po_box(s):
        """
        Part of text_formatter. Standardized po box text to PO BOX
        @param s: string
        @return: string
        """
        s = RegexHandler.po_box1.sub(r'\nPO BOX \1', s)
        return RegexHandler.po_box2.sub(r'PO BOX \1', s)

    @staticmethod
    def sub_no_number(s):
        """
        Part of text_formatter. handles "  no:/number: 123456" -> "  123456"
        @param s: string
        @return: string
        """
        return RegexHandler.no_number.sub(r'\1 \3', s)

    @staticmethod
    def sub_patterns_2_tel(s):
        """
        Part of text_formatter. Replaces several telephone prefix to "Tel: "
        @param s: string
        @return: string
        """
        text = RegexHandler.direct.sub(r'Direct: \2', s)
        return RegexHandler.tel.sub(r'Tel:  \2', text)

    @staticmethod
    def sub_toll_free(s):
        """
        Part of text_formatter. Replaces several Toll-free  to "Toll-Free: "
        @param s: string
        @return: string
        """
        return RegexHandler.toll_free.sub(r'Toll-Free', s)

    @staticmethod
    def remove_non_printable_chars(s):
        """Replace non-printable characters in a string."""
        # the translate method on str removes characters
        # that map to None from the string
        return s.translate(RegexHandler.NOPRINT_TRANS_TABLE)

    @staticmethod
    def condense_numbers(t):
        """
        every two numbers separated by "," are condensed together
        @param t:
        @return: t
        """
        return re.sub(r'(\d+),(\d+)', r'\1\2', t)

    # @staticmethod NOT USED
    # def contain_usd_amounts(t):
    #     return len(RegexHandler.usd_amount.findall(t)) > 0

    # @staticmethod NOT USED
    # def get_usd_amounts(t):
    #     return RegexHandler.usd_amount.findall(t)

    # @staticmethod NOT USED
    # def set_keywords(t_list):
    #     if type(t_list) is not list:
    #         t_list = [t_list]
    #     RegexHandler.keywords = [t.lower() for t in t_list]

    # @staticmethod NOT USED
    # def add_keyword(t):
    #     RegexHandler.keywords.append(t.lower())

    @staticmethod
    def is_keyword(t):
        return bool(t.lower() in RegexHandler.keywords)

    # @staticmethod NOT USED
    # def get_keyword_id(t):
    #     return 0 if t.lower() not in RegexHandler.keywords else (RegexHandler.keywords.index(t.lower()) + 1)

    @staticmethod
    def get_emails(t, unique=True):
        """
        finds all email patterns in string
        @param t: string
        @param unique: bool
        @return: list of emails
        """
        emails = RegexHandler.email.findall(t)
        if unique:
            return list(sorted(set(emails)))
        return emails

    # @staticmethod NOT USED
    # def count_substr_in_str(substr, full_str):
    #     count = sum(1 for _ in re.finditer(r'\b%s\b' % re.escape(substr), full_str))
    #     return count

    @staticmethod
    def get_domains_and_users(text, emails=None, return_emails=False):
        """
        Extract domain & user (first/last name) in a text (signature).
        we assume that the domain is in the email pattern and also in the urls
        and the user is in the email pattern
        @param text: string
        @param emails:
        @param return_emails:
        @return: list of domain, users and if needed also the emails
        """
        if emails is None:
            emails = []
        if not emails or type(emails) is not list:
            emails = RegexHandler.get_emails(str(text), unique=True)
        domains = []
        users = []
        for email in emails:
            domain = '.'.join(email.split('@')[1].split('.')[0:-1])
            domain = domain.split('.')[0]
            user = email.split('@')[0]
            domains.append(domain)
            users.append(user)
        if return_emails:
            return list(sorted(set((domains)))), list(sorted(set((users)))), list(sorted(set((emails))))
        return list(sorted(set((domains)))), list(sorted(set((users))))

    @staticmethod
    def get_domains(t, unique=True, no_social=False, emails=True, urls=True):
        """
        Extract the domain of the company from emails / urls in a string
        @param t: string (Text)
        @param unique: bool
        @param no_social: bool
        @param emails: list
        @param urls: list
        @return: list of domains
        """
        domains = []
        if emails:
            domains = ['.'.join(email.split('@')[1].split('.')[0:-1]) for email in RegexHandler.get_emails(t, unique)]
        if urls:
            urls = RegexHandler.get_urls(t, True)
            try:
                url_domains = [tldextract.extract(url).domain for url in urls]
                domains =  url_domains+domains
            except:
                for url in urls:
                    url_split = url.split('.')
                    if len(url_split) > 2:
                        if '.co.' in url:
                            domains += [url_split[-3]]
                        else:
                            domains += [url_split[-2]]
                    elif len(url_split) == 2:
                        if 'www' not in url_split[0].lower():
                            domains += [url_split[0]]
        if unique:
            domains = list(sorted(set(domains)))
        if no_social:
            social_domains = ['edgepilot', 'bit.ly', 'linkedin', 'goo', 'google', 'facebook', 'skype', 'pintrest',
                              'youtube', 'tinyurl', 'youtu.be']
            domains = list(filter(lambda domain: domain not in social_domains, domains))
        return domains

    @staticmethod
    def get_urls(t, unique=True):
        """
        Extracts the urls in a text (http://, www. etc.)
        @param t: Text
        @param unique: bool
        @return: list of urls
        """
        url = RegexHandler.url2.findall(t)
        urls = [x[0] for x in url]
        if unique:
            urls = list(sorted(set(urls)))
        return urls

    # @staticmethod NOT USED
    # def get_urls_old(t, unique=True):
    #     urls = RegexHandler.url.findall(t)
    #     if unique:
    #         z = []
    #         for ul in urls:
    #             ls = [ll for ll in list(sorted(set((ul))) if ll]
    #             z += ls
    #         return list(sorted(set((z)))
    #     return urls

    @staticmethod
    def get_emojis(t, unique=True):
        """
        Extract emojis in text
        @param t: text
        @param unique: bool
        @return: list of emojis
        """
        emojis = RegexHandler.emoji.findall(t)
        if unique:
            return list(sorted(set(emojis)))
        return emojis

    @staticmethod
    def remove_emojies(t):
        """
        removes emojis found in text
        @param t: text
        @return: text without emojis
        """
        return RegexHandler.emoji.sub(r'', t)

    @staticmethod
    def get_numbers(t, unique=True):
        """
        get all numbers in text
        @param t: text
        @param unique: bool
        @return: list of numbers in text
        """
        numbers = RegexHandler.number.findall(t)
        if unique:
            return list(sorted(set(numbers)))
        return numbers

    @staticmethod
    def contain_email(t):
        """
        checks if text contains an email pattern
        @param t: text
        @return: bool
        """
        return len(RegexHandler.get_emails(t)) > 0

    @staticmethod
    def contain_url(t):
        """
        checks if text contains a url pattern
        @param t: text
        @return: bool
        """
        return len(RegexHandler.get_urls(t)) > 0

    @staticmethod
    def contain_emoji(t):
        """
        checks if text contains an emoji
        @param t: text
        @return: bool
        """
        return len(RegexHandler.get_emojis(t)) > 0

    @staticmethod
    def contain_number(t):
        """
        checks if text contains a number
        @param t:
        @return: bool
        """
        return len(RegexHandler.get_numbers(t)) > 0

    @staticmethod
    def get_number_of_digits_in_number(n):
        """! This function return the number of digits in number
         @verbatim
         Args:
            n : Number
         Returns:
            digits: # of digits in the number.
         @endverbatim
         """
        if n == '':
            return 0
        if type(n) is str:
            return len(n)
        if n > 0:
            digits = int(math.log10(n)) + 1
        elif n == 0:
            digits = 1
        else:
            digits = int(math.log10(-n)) + 2  # +1 if you don't count the
        return digits

    @staticmethod
    def get_number_of_digits_and_digits_groups_in_text(t):
        """
        get all numbers and their number of digits in text
        @param t:
        @return: list of numbers and their digit amount
        """
        groups = []
        n_digits = 0
        group = ''
        for c in t:
            if c.isdigit():
                n_digits += 1
                group += c
            else:
                if len(group):
                    groups.append(group)
                    group = ''
        if len(group):
            groups.append(group)
        return n_digits, groups

    @staticmethod
    def get_users(text):
        """
        splitting the user from the email pattern and get all users from the emails
        @param text: string
        @return: list of users
        """
        emails = RegexHandler.get_emails(text, unique=True)
        users = []
        for email in emails:
            user = email.split('@')[0]
            users.append(user)
        return list(sorted(set((users))))

    # noinspection PyPep8Naming
    @staticmethod
    def now_as_string():
        """
        getting current date time as string in the format bellow
        @return: current datetime as string
        """
        currentDT = datetime.datetime.now()
        return currentDT.strftime("%a, %b %d, %Y at %I:%M:%S %p")

    @staticmethod
    def contain_month(t):
        """
        checks if text contains one of the calender months - january-december
        @param t: text
        @return: bool
        """
        for tt in RegexHandler.remove_punct(t.lower()).split():
            if tt in RegexHandler.months:
                return True
        return False

    @staticmethod
    def contain_day(t):
        """
        checks if text contains one of the weekly days - sunday-saturday
        @param t: text
        @return: bool
        """
        for tt in RegexHandler.remove_punct(t.lower()).split():
            if tt in RegexHandler.days:
                return True
        return False

    @staticmethod
    def is_time_format(t):
        """! Is time format is valid
          @verbatim
          Args:
             t : time
          Returns:
             bool : TRUE if time match the format
          @endverbatim
        """
        return bool(RegexHandler.time.match(t))

    @staticmethod 
    def has_time_format(t):
        """! Has time format.
          @verbatim
          Args:
             t : time
          Returns:
             bool : if input t, has the time format correctly - > return True.
          @endverbatim
        """
        t = t.lower().replace('am', ' ').replace('pm', ' ')
        r_list = RegexHandler.has_time.findall(t)
        if not len(r_list):
            return False
        else:
            for l in r_list:
                h = int(l[0])
                m = int(l[1])
                if h not in range(0, 25) or m not in range(0, 61):
                    return False
            return True

    @staticmethod
    def remove_html(t):
        """! Removing HTML template.
          @verbatim
          Args:
             t : text
          Returns:
             string : without html template
          @endverbatim
        """
        return RegexHandler.html.sub(r'', t)

    @staticmethod
    def merge_multiple_spaces(t):
        """
        turns multiple spaces to one
        @param t: text
        @return: text
        """
        return RegexHandler.merge_spaces.sub(r' ', t)

    @staticmethod
    def merge_multiple_new_lines(t):
        """! Merge multiple new lines.
          @verbatim
          Args:
             t : Text
          Returns:
             new text : merged lines (removed the \n ).
          @endverbatim
          """
        return RegexHandler.merge_lines.sub('\n', t).strip()

    @staticmethod
    def contain_camel_case(t):
        num_tokens_before_split_on_case = len(t.split())
        camel_case_splitted = RegexHandler.camelCase1.sub(r' \1', RegexHandler.camelCase2.sub(r' \1', t)).split()
        num_tokens_after_split_on_case = len(camel_case_splitted)
        return num_tokens_after_split_on_case > num_tokens_before_split_on_case

    @staticmethod
    def camel_to_snake(t):
        s1 = RegexHandler.camelCase3.sub(r'\1_\2', t)
        s2 = RegexHandler.camelCase4.sub(r'\1_\2', s1)
        if '_' in s2:
            # check if we have several UPPERS - if so split differently
            s3 = RegexHandler.camelCase5.sub(r'\1\2_', s2)
            return s3
        return s2

    @staticmethod
    def is_hashtag(t, edit=False):
        if t.startswith('#'):
            return True if not edit else (True, t.replace('#', ''))
        return False if not edit else (False, t)

    @staticmethod
    def is_user(t, edit=False):
        if t.startswith('@'):
            return True if not edit else (True, t.replace('@', ''))
        return False if not edit else (False, t)

    @staticmethod
    def contain_non_ascii_chars(t, edit=False, return_only_ascii=False):
        """
        counts the ascii, non ascii chars in the text.
        depend on the arguments, it can return:
            1. count of non ascii
            2. only the ascii chars
            3. count of non ascii chars, ascii chars, non ascii chars
        @param t:text
        @param edit: bool
        @param return_only_ascii: bool
        @return: decribed above
        """
        if not edit:
            return sum([(ord(s) >= 128) * 1 for s in t])
        asciis = ''
        non_asciis = ''
        for s in t:
            if ord(s) >= 128:
                non_asciis += s
            else:
                asciis += s
        if return_only_ascii:
            return asciis
        return len(non_asciis), asciis, non_asciis

    @staticmethod
    def remove_punct(text='', right=True, left=True, middle=True):
        """
        removes all punctuations from token (",",".","-", etc)
        by default it will remove all. in case given a specific direction, it
        start from it until reaches a non punctuations char
        @param text: text
        @param right: start from right side
        @param left: start from left side
        @param middle: do not modify most left and right chars
        @return: string
        """
        if right and left and middle:
            return text.translate(str.maketrans('', '', punctuation))

        token = text
        if left:
            while token and token[0] in punctuation:
                token = token[1:] if len(token) > 1 else ''

        if right:
            while token and token[-1] in punctuation:
                token = token[:-1] if len(token) > 1 else ''

        if middle and len(token) > 2:
            token = token[0] + token.translate(str.maketrans('', '', punctuation)) + token[-1]

        return token

    @staticmethod
    def remove_non_ascii_chars(t, right=True, left=True, middle=True):
        """

        removes all non ascii from token (",",".","-", etc)
        by default it will remove all. in case given a specific direction, it
        start from it until reaches a non ascii char
        @param text: text
        @param right: start from right side
        @param left: start from left side
        @param middle: do not modify most left and right chars
        @return: string
        """
        if right and left and middle:
            return ''.join([s for s in t if ord(s) < 128])
        else:
            if left:
                while t and ord(t[0]) >= 128:
                    t = t[1:] if len(t) > 1 else ''
            if right:
                while t and ord(t[-1]) >= 128:
                    t = t[:-1] if len(t) > 1 else ''
            if middle and len(t) > 2:
                t = t[0] + RegexHandler.remove_non_ascii_chars(t[1:-1]) + t[-1]
            return t

    @staticmethod
    def is_bad_decoding_line(line):
        """! Return if this line has bad decoding.
        @verbatim
        Args:
           line : Line
        Returns:
           bool : True if this line has bad decoding.
        @endverbatim
        """
        return len(re.findall(RegexHandler.bad_decoding, line, flags=0)) > 0

    @staticmethod
    def contain_date(tt):
        """! Check if the text contains date.
        @verbatim
        Args:
           tt : text we want to check if it's date.
        Returns:
           bool : True if contain date.
        @endverbatim
        """
        def remove_punct(s):
            return " ".join(s.translate(str.maketrans('', '', string.punctuation)).split()).strip()
    
        for t in tt.split():
            for marker in ['.', '|', '/', '-']:
                if marker in t:
                    if remove_punct(t).isdigit():
                        template = "%m/%d/%y" if '/' in t else "%m|%d|%y" if '|' in t else "%m-%d-%y" if '-' in t else "%m.%d.%y"
                        try:
                            datetime.datetime.strptime(t, template)
                            return True
                        except:
                            continue
                    continue
        return False

    @staticmethod
    def remove_brackets_content_from_email_text(text):
        # t = text.replace('[mailto:', '')
        #             t = RegexHandler.square_brackets.sub('', t)
        t = text.replace('[', '').replace(']', '')
        t = RegexHandler.brackets_cid.sub('', t)
        return t

    @staticmethod
    def extract_emojies_codes(input_string, check_first=True):
        """! Remove duplicates lines.
        @verbatim
        Args:
           input_string : String
           check_first(=True): Check input string if contains an emoji.
        Returns:
            returnString, emojies, ascies
        @endverbatim
        """
        if check_first:
            if not RegexHandler.contain_emoji(input_string):
                return input_string
    
        returnString = ""
        emojies = []
        ascies = []
        for character in input_string:
            try:
                character.encode("ascii")
                returnString += character
            except UnicodeEncodeError:
                replaced = unidecode(str(character))
                if replaced != '':
                    returnString += replaced
                    ascies += [character]
                else:
                    try:
                        emojies += [unicodedata.name(character).replace(' ', '_')]
                    except ValueError:
                        ascies += [character]
    
        return returnString, emojies, ascies

    @staticmethod
    def normalize_space(text):
        """
        Remove multiple space characters and keep single simple space. Also stripping new lines ending the text input.
        :param text: string input
        :return: processed string
        """
        return '\n'.join([' '.join(line.split()) for line in text.strip().split('\n')])

    # noinspection RegExpSingleCharAlternation
    @staticmethod
    def remove_space_before_semicolon(text):
        """
        Replace any non alphanumeric character preceded and followed by a space, with the character without space
        preceding. For example: 'Mobile : 972-54654' will output: 'Mobile: 972-54654'
        :param text: string input
        :return: processed string
        """
        # return re.sub(r'\s+(:|,)', r'\1 ', text)
        return RegexHandler.semicolon_spaces.sub(r'\1 ', text)

    @staticmethod
    def remove_brackets_for_digits(text):
        """
        Remove brackets around one or more digits that may be preceded by a plus sign (+). The output is the digits
        with the plus sign if present without the beackets.
        :param text: string input
        :return: processed string
        """
        # return re.sub(r'\(\s?(\+?\d+)\s?\)', r'\1', text)
        return RegexHandler.digits_brackets.sub(r'\1',text)

    @staticmethod
    def count_sequential_digits(text, get_all=False):
        """
        fetches all numbers in text, and depends on the get_all argument it returns either the
        max length number, or the length of all numbers
        @param text: text
        @param get_all: bool
        @return: list of lengths
        """
        # matches = re.findall(r'\d[\d\s\-.]+\d', text)
        matches = re.findall(r'\d+', RegexHandler.remove_punct(text))
        if matches:
            if not get_all:
                return max([len([m for m in match if m.isdigit()]) for match in matches])
            else:
                return [len([m for m in match if m.isdigit()]) for match in matches]
        return 0 if not get_all else [0]

    # noinspection RegExpSingleCharAlternation
    @staticmethod
    def insert_space(text):
        """
        replace chars that represents a separation with whitespace
        @param text: text
        @return: text
        """
        # return re.sub(r'(\*|~|\|)', r' \1 ', text)
        return RegexHandler.rep_with_space.sub(r' \1 ', text)

    @staticmethod
    def split_x_number(text):
        """
        in case of extentions that appear as "x", we split the number and the "x"
        @param text:  text
        @return: text
        """
        # return re.sub(r'\bx(\d+)\b', r'x \1', text)
        return RegexHandler.number_with_x.sub(r'x \1',text)

    @staticmethod
    def split_number_non_number(text):
        """
        if we have this pattern phone-123456, we want to split the "phone" from the nymber itself
        we have a special treatment for linkedin profiles that do have such pattern and we keep them as is
        @param text: text
        @return:text
        """
        # don't separate linkedin profile numbers
        text = re.sub(r'(linkedin\.com/in/)([\w]+-[\w]+)-(\d{6,20})', r'\1\2_\3', text)
        text = re.sub(r'\b([a-zA-Z]+)-(\d{6,20})\b', r'\1 \2', text)
        text = re.sub(r'(linkedin\.com/in/)([\w]+-[\w]+)_(\d{6,20})', r'\1\2-\3', text)
        return text  # re.sub(r'\b([a-zA-Z]+)-(\d{6,20})\b', r'\1 \2', text)

    # noinspection RegExpRedundantEscape
    @staticmethod
    def concat_numbers(text):
        """
        when there are two phone/mobile numbers separated with white space, we will add ","
        between them for easier preprocessing
        @param text: text
        @return: text
        """
        # text = re.sub(r'(\d{6,20}/\d*)\s+(\d{8,20})', r'\1, \2', text)
        # text = re.sub(r'(\d{5,20})\s+(\d{3}-\d+\D*)', r'\1, \2', text)
        # text = re.sub(r'(\d{3,4}-\d{3,4})\s+(\d+-\d+)', r'\1, \2', text)
        # text = re.sub(r'(\d{6,20})\s+(\d{5,20})', r'\1, \2', text)
        text = RegexHandler.concat_phones1.sub(r'\1, \2', text)
        text = RegexHandler.concat_phones2.sub(r'\1, \2', text)
        text = RegexHandler.concat_phones3.sub(r'\1, \2', text)
        text = RegexHandler.concat_phones4.sub(r'\1, \2', text)

        # text = re.sub(r'(?<=\d) (?=\d)', '-', text)
        text = re.sub(r'(?<=[^a-zA-Z][\d+\-\(\)]) (?=[\d\-\(\)][^a-zA-Z])', '-', text)  # dorons fix
        # text = '\n'.join([re.sub(r'(?<=\d)+\s+(?=\d+(\s|\b))', '-', line) for line in text.split('\n')])
        # text = re.sub(r'([\s-]\d+)( )(\d+\s)', r'\1-\3', text)
        # text = re.sub(r'([\s-]\d+)( )(\d+\s)', r'\1-\3', text)
        # text = re.sub(r'(^\d+)( )(\d+[\s-])', r'\1-\3', text)
        # text = re.sub(r'([\s-]\d+)( )(\d+$)', r'\1-\3', text)

        return text

    @staticmethod
    def fix_url(text):
        """
        adds www to token that contains a know domain for later use in connections handler
        @param text: text
        @return: text
        """
        line_l = text.split('\n')
        line_l_out = []
        for line in line_l:
            token_l = line.split(' ')
            token_l_out = []
            for token in token_l:
                if '@' not in token and re.findall(r'\S+(\.org|\.gov|\.com|\.net|\.int|\.edu|\.mil)(\s|$)', token):
                    if not token.startswith('www.'):
                        token_l_out.append('www.' + token)
                    else:
                        token_l_out.append(token)
                else:
                    token_l_out.append(token)
            line_l_out.append(' '.join(token_l_out))
        return '\n'.join(line_l_out)

    @staticmethod
    def clear_url_junk(text):
        """
        detect patterns of url junk such %3A%2F%2F and removes it
        @param text: text
        @return: text
        """
        token_l = text.split()
        token_l_out = []
        for token in token_l:
            if token.count('%') > 1:
                if len(RegexHandler.url_junk.findall(token)) == len(token):
                    continue
                else:
                    token_l_out.append(token)
            else:
                token_l_out.append(token)
        return ' '.join(token_l_out)

    @staticmethod
    def clear_contact_prefixes(text):
        """
        removes the contact prefix (emial:, linkedin:, twittwer, etc)
        @param text: text
        @return: text
        """
        # return re.sub(RegexHandler.contact_prefixes_pattern, '', text, flags=re.IGNORECASE)
        return RegexHandler.contact_prefixes.sub('', text)

    # noinspection RegExpRedundantEscape
    @staticmethod
    def clear_number_suffixes(text):
        """
        TODO daniela add description of this regex
        @param text:
        @return:
        """
        return re.sub(r'([\d\-\(\)\.\+]{5})_([a-zA-Z]+)($|\s)', r'\1 \2\3', text)

    @staticmethod
    def get_digits(text):
        """
        Extract only digits from number
        @param text: string to process
        @return: ordered digits in string
        """
        return ''.join(re.findall(r'\d', text))

    # @staticmethod
    # def split_line_by_separator(text):
    #     split_sep = RegexHandler.split_by_separator(text)
    #     if len(split_sep) == 2:
    #         if len(split_sep) >= 2:
    #             return '\n'.join(split_sep)
    #     return text
    #
    # # noinspection RegExpRedundantEscape
    # @staticmethod
    # def split_by_separator(text):
    #     return RegexHandler.split_by_separator.split(text)



#!/usr/bin/env python
# coding: utf-8

"""! @brief ZIDS_Pipeline base class for the pipeline."""

from typing import List
from dsframework.base.common.component import ZIDS_Component
from dsframework.base.pipeline.predictables.predictable import ZIDS_Predictable
from dsframework.base.pipeline.artifacts.shared_artifacts import ZIDS_SharedArtifacts


class ZIDS_Pipeline():
    """! ZIDS_Pipeline base class for the pipeline.

    Its main job is the base for building pipeline components.
    """

    def __init__(self):
        """! ZIDS_Pipeline class initializer."""
        self.components:List[ZIDS_Component] = []
        self.predictables:List[ZIDS_Predictable] = []
        self.artifacts = self.get_artifacts()
        self.build_pipeline()
        self.configure_pipeline()
        self.connect_pipeline()

    def get_artifacts(self):
        """! Loads and returns the artifacts and vocabs,

        Overridden by its parent in generatedProjectNamePipeline.get_artifacts() method."""
        return ZIDS_SharedArtifacts()
    
    def build_pipeline(self):
        """! Main place where the pipeline gets built with all of its components.

        Overridden by default implementation in generatedProjectNamePipeline.build_pipeline(), where its
        four main components gets instantiated:

        - generatedProjectNamePreprocess
        - generatedProjectNamePredictor
        - generatedProjectNameForcer
        - generatedProjectNamePostprocess.
        """
        raise NotImplementedError

    def configure_pipeline(self, **kwargs):
        """! Add configurations that need to take place after the build_pipeline() method.

        Override method in generatedProjectNamePipeline.configure_pipeline()
        """
        pass

    def connect_pipeline(self):
        """! Method distributes Artifacts instance to all pipeline components."""
        for c in self.components:
            c.artifacts = self.artifacts

    def preprocess(self, **kwargs) -> List[ZIDS_Predictable]:
        """! Runs preprocess step in the beginning of the pipeline

        Override method in generatedProjectNamePipeline.preprocess(), it needs to include all required steps
        before the Predictor (model) step.

        Args:
            **kwargs: Dataset loaded initially.
        Returns:
            List[ZIDS_Predictable]: List of predictable objects."""
        raise NotImplementedError

    def postprocess(self, predictables):
        """! Runs postprocess step at the end of the pipeline.

        Override method in generatedProjectNamePipeline.postprocess(), it needs to return the list of results.

        Returns:
            List[generatedProjectNameOutputs]: List of results.
        """
        raise NotImplementedError


    def add_component(self, component:ZIDS_Component):
        """! Adds component to pipeline.

        There are two components added by default: Predictor, Forcer and this is in addition to the pre-existing ones
        the preprocessor and postprocessor.

        Args:
            component: ZIDS_Component, component to add.
        """
        self.components.append(component)

    def __call__(self,  **kwargs):
        """! ZIDS_Pipeline __call__() method, runs execute() method of this class with specified args.

        Args:
            **kwargs: Initially loaded dataset.
        """
        return self.execute( **kwargs)

    def execute(self, **kwargs):
        """! Executes the pipeline,

        Runs the execute method for all registered components one after the other,

        Args:
            **kwargs: Initially loaded dataset.
        """
        self.predictables = self.preprocess(**kwargs)
        for c in self.components:
            self.predictables = c.execute(self.predictables)
        return self.postprocess(self.predictables)



registry = {}


class MultiMethod(object):
    def __init__(self, name):
        self.name = name
        self.typemap = {}

    def __call__(self, *args):
        types = tuple(arg.__class__ for arg in args)  # a generator expression!
        function = self.typemap.get(types)
        if function is None:
            raise TypeError("no match")
        return function(*args)

    def register(self, types, function):
        if types in self.typemap:
            raise TypeError("duplicate registration")
        self.typemap[types] = function



#!/usr/bin/env python
# coding: utf-8

##
# @file
# @brief ZIDS_Preprocessor base class for Preprocess class.

from typing import List, Any
from dsframework.base.common.component import ZIDS_Component
from dsframework.base.pipeline.predictables.predictable import ZIDS_Predictable
from dsframework.base.pipeline.artifacts.shared_artifacts import ZIDS_SharedArtifacts

class ZIDS_Preprocessor(ZIDS_Component):
    """! ZIDS_Preprocessor the base class for Preprocess class.

    This is the first step of the pipeline and its main goal is to prepare our dataset format to the model.
    """

    def __init__(self, artifacts:ZIDS_SharedArtifacts=None):
        """! ZIDS_Preprocessor initializer

        Args:
            artifacts(ZIDS_SharedArtifacts): Shared artifacts instance.
        """

        ##
        # @hidecallgraph @hidecallergraph
        super().__init__(artifacts)
        self.input = None

    def __call__(self, **kwargs: Any) -> Any:
        """! ZIDS_Preprocessor.preprocess __call__() method gets called when its instance gets called ().

        Args:
            **kwargs: Data loaded initially to run model on.
        Returns:
            List[ZIDS_Predictable]: Result of execute method with **kwargs as parameter.
        """
        return self.execute(**kwargs)

    def normalize_input(self, **kwargs: Any) -> Any:
        """! Converts input to Inputs datatype.

        Empty base method, implemented in generatedProjectNamePreprocess class."""
        raise NotImplementedError

    def preprocess(self, raw_input:Any):
        """! Preprocess step, add preprocess procedure here.

        Method not implement, override in generatedProjectNamePreprocess class and return a list of predictable objects.

        Args:
            raw_input: Data loaded initially to run model on.

        Returns:
            List[ZIDS_Predictable] - Not implemented yet.

        Raises:
            NotImplementedError

        """
        raise NotImplementedError
        ## Should return a list of predictable objects

    def execute(self, **kwargs: Any) -> List[ZIDS_Predictable]:
        """! Executes preprocess step.

        Args:
            **kwargs: Data loaded initially to run model on.
        Returns:
            List[ZIDS_Predictable]: List of predictable objects
        """
        self.input = self.normalize_input(**kwargs)
        return self.preprocess(self.input)




from pandas import DataFrame
from pydantic import BaseModel

##
# @file
# @brief Includes Evaluator base class for local and cloud evaluation, ConfusionMatrix class and ConfusionRow to hold
# evaluation result.

class ZIDS_Evaluator():
    """ZIDS_Evaluator base class for evaluator.py, it is used in local and cloud evaluation."""
    def __init__(self):
        """ZIDS_Evaluator initializer"""
        pass

    def evaluate_model_predictions(self, df_from_model: DataFrame) -> DataFrame:
        """! Apply evaluation function (_get_confusion_matrix) on every incoming Dataframe row.

        Args:
            df_from_model: A dataframe containing both labeled data and model predictions
        Returns:
            Altered dataframe with a new column 'confusion_matrix' that holds a list of 'ConfusionRow' objects
        """
        df_from_model['confusion_matrix'] = df_from_model.apply(lambda row:
                                                                self.get_confusion_matrix(row.pred, row.pred_model),
                                                                axis=1)

        return df_from_model

    def get_confusion_matrix(self, true_label, prediction):
        """Model evaluation function, to be implemented per project in evaluator.py file"""
        raise NotImplementedError

    @staticmethod
    def get_confusion_matrix_binary(true_label, prediction):
        """! Gets a single confusion matrix, in a form of a dictionary:
        @code {.py}
        confusion_matrix = {'tp': 0, 'fp': 0, 'fn': 0, 'tn': 0}
        @endcode
        """
        confusion_matrix = {'tp': 0, 'fp': 0, 'fn': 0, 'tn': 0}
        if true_label == 1 and prediction == 1:
            confusion_matrix['tp'] = 1
        elif true_label == 1 and prediction == 0:
            confusion_matrix['fn'] = 1
        elif true_label == 0 and prediction == 0:
            confusion_matrix['tn'] = 1
        else:  # true_label == 0 and prediction == 1
            confusion_matrix['fp'] = 1
        return {'cm': confusion_matrix}


class ConfusionMatrix(BaseModel):
    """Confusion matrix class, used as a base class for ConfusionRow, holds the evaluation results and
    provides methods to calculate required matrix parameters, such as precision, recall and f1."""
    def __init__(self):
        self.tp: int = 0  # True Positive
        self.tn: int = 0  # True Negative
        self.fp: int = 0  # False Positive
        self.fn: int = 0  # False Negative

        self.precision: float = 0
        self.f1: float = 0
        self.recall: float = 0

    def __add__(self, other):
        """Add results to confusion matrix"""
        _sum = ConfusionMatrix(
            tp=self.tp + other.tp,
            fp=self.fp + other.fp,
            fn=self.fn + other.fn,
            tn=self.tn + other.tn,
        )
        _sum.calc()
        return _sum

    def precision_calc(self):
        """Calculate precision parameter: true_positive / (true_positive + false_positive) """
        sum_ = self.tp + self.fp
        if sum_:
            return self.tp / sum_
        else:
            return 0

    def recall_calc(self):
        """Calculate recall parameter: true_positive / (true_positive + false_negative) """
        sum_ = self.tp + self.fn
        if sum_:
            return self.tp / sum_
        else:
            return 0

    def f1_calc(self):
        """Calculate f1 parameter: 2 * (precision * recall) / (precision + recall) """
        precision_ = self.precision
        recall_ = self.recall
        if precision_ + recall_ > 0:
            return 2 * (precision_ * recall_) / (precision_ + recall_)
        return 0

    def calc(self):
        """Calls precision, recall and f1 parameters calculations."""

        ## Precision parameter value
        self.precision= round(self.precision_calc(),3)
        ## Recall parameter value
        self.recall = round(self.recall_calc(), 3)
        ## f1 parameter value
        self.f1 = round(self.f1_calc(), 3)


class ConfusionRow(ConfusionMatrix):
    """
    This class holds information regarding evaluation result of a single attribute that was sent for evaluation.
    It inherits from the ConfusionMatrix class to enable confusion matrix abilities, and will eventually output
    all relevant values for a single attribute. The evaluator output should be a list of objects from this class type
    """

    def __init__(self, name: str = "generic_cm"):
        """ConfusionRow initializer."""

        ##
        # @hidecallgraph @hidecallergraph

        super().__init__()
        ## Assigns a name, defaults name: str = "generic_cm"
        self.name = name
        self.pred_val: list = []
        self.target_val: list = []

    def set_val_str(self, target: [str, None], pred: [str, None]):
        """! Evaluates a string value and updates confusion matrix.

        Args:
            target: Ground truth
            pred: Current prediction.

        Raises:
            ValueError: On wrong input values.
        """
        # pred='' and target=''
        if pred in [None, ''] and target in [None, '']:
            self.tn += 1
        # pred='' and target='sdfsdf'
        elif pred in [None, ''] and target not in [None, '']:
            self.fn += 1
        # ('a'!='b') or (pred='sdfsdf' & target ='')
        elif pred != target:
            self.fp += 1
        elif pred == target:
            self.tp += 1

        else:
            raise ValueError

        self.pred_val.append(pred)
        self.target_val.append(target)
        self.calc()

    def set_val_int(self, target: int, pred: int, zero_is_empty=False):
        """! Evaluates an integer value and updates confusion matrix.

        Args:
            target: Ground truth
            pred: Current prediction.
            zero_is_empty: Default false - adds values to pred_val and target_val.

        """
        if pred == target:
            if (pred != 0) or (zero_is_empty is False):
                self.tp += 1
        elif pred > target:
            self.fp += 1
        elif target > pred:
            self.fn += 1

        if (zero_is_empty is False) or (pred != 0 or target != 0):
            self.pred_val.append(pred)
            self.target_val.append(target)
        self.calc()

    def set_val_list_count(self, target: [int, None], pred: [int, None]):
        """! Evaluates an integer value and updates confusion matrix.

        Args:
            target: Ground truth
            pred: Current prediction.

        """
        if pred is None and target is not None:
            self.fn += 1
        elif pred is not None and target is None:
            self.fp += 1
        elif pred == target:
            self.tp += target
        elif pred > target:
            self.fp += 1
        elif target > pred:
            self.fn += 1

        self.pred_val.append(pred)
        self.target_val.append(target)
        self.calc()

    def set_val_check_list_of_list(self, col: list, exp: list, pre_comp_func_l=[]):
        """! Compare two list and count how many values UNIQUE from each appears at the other.

        Args:
            col: List 1
            exp: List 2
            pre_comp_func_l: Not used
        """
        ##
        # examples:
        # @code{.py}
        # miss col - Exp: [A, B, C] Col: [B,C] tp:2 fn:1
        # miss exp - Exp: [A] Col: [A, B, C] tp:1 fp: 2
        # @endcode
        # The Q unique or not unique:
        # @code{.py}
        # unique     - Exp: [A,A,A] col [A,B,B] tp:1 fp:1
        # unique     - Exp: [A,A,A,A] col [A,B,B] tp:1 fp:1
        # @endcode
        # or
        # @code{.py}
        # not unique - Exp: [A,A,A] col [A, B] tp:1 fp: 1 fn: 1
        # not unique - Exp: [A,A,A,A] col [A,B] tp:1 fp:1 fn :2
        # tmp_pred = copy.deepcopy(pred)
        # tmp_target = copy.deepcopy(target)
        #
        # for func in pre_comp_func_l:
        #     tmp_pred = func(record_l=tmp_pred)
        #     tmp_target = func(record_l=tmp_target)
        # @endcode

        tmp_col = col
        tmp_exp = exp

        exp_idx = 0
        for exp_idx, exp in enumerate(tmp_exp):
            if exp_idx < len(tmp_col):
                if exp == tmp_col[exp_idx]:
                    self.tp += 1
                else:
                    # found values on pred and target that mismatch
                    self.fp += 1
            else:
                # no value at predict
                self.fn += 1

        if exp_idx + 1 < len(tmp_col):
            # adiel: tn used to mark error in this case YAK!
            self.fp += len(tmp_col) - (exp_idx + 1)
        self.calc()

    def set_exp_col_cm(self, match_num, exp_miss, col_miss):
        """! Set results to self confusion matrix.

        Args:
            match_num: Sets true_positive
            exp_miss: Sets false negative
            col_miss: Sets false positive.

        """

        # row.set_exp_col_cm(match_num=len(order_l['match_exp']), exp_miss=len(order_l['exp_no']), col_miss=len(order_l['col_no']))
        ## True positive
        self.tp = match_num
        ## False negative
        self.fn = exp_miss
        ## False positive
        self.fp = col_miss
        self.calc()

    def get_cols_per_attr(attr):
        """! Gets a list of attribute strings per confusion matrix value.

        Returns:
            [f'{attr}_pred', f'{attr}_target', f'{attr}_prob', f'{attr}_tp', f'{attr}_fp', f'{attr}_tn', f'{attr}_fn']
        """
        return [f'{attr}_pred', f'{attr}_target', f'{attr}_prob', f'{attr}_tp', f'{attr}_fp', f'{attr}_tn',
                f'{attr}_fn']

    def get_server_tester(self):
        """! Constructs a dictionary with confusion matrix, pred, prob and target results.

        Returns:
            Dictionary with results.
        """
        self.calc()
        result = {
            self.name + '_fn': self.fn,
            self.name + '_fp': self.fp,
            self.name + '_tn': self.tn,
            self.name + '_tp': self.tp,
            self.name + '_pred': str(self.pred_val),
            self.name + '_prob': -1,
            self.name + '_target': str(self.target_val),
        }
        return result



import os
from json import load as json_load
from json import dumps as json_dumps
from google.protobuf.duration_pb2 import Duration
from google.cloud.dataproc_v1 import ClusterControllerClient, Cluster, ClusterConfig
from google.cloud import storage
import subprocess
import sys


##
# @file
# @brief Helper functions for batch projects.


class ZIDSBatchHelper:
    """! General class for a variety of helper functions for batch processing projects
    The capabilities of this class include:
    *) Parsing batch configuration file
    *) Creating a Dataproc Cluster
    *) Uploading jars, stages and wheel archive to a given bucket
    """

    def __init__(self, stage_config):
        """! Initialize the class, primarily fill out configuration parameters

            Args:
                stage_config : Configuration dictionary, loaded from configuration file.
        """
        # Save the entire config file for cluster creation
        self.config = stage_config

        self.user_email = ''
        env_type = os.environ.get('SPRING_PROFILES_ACTIVE')
        if (not env_type == 'production') and (not env_type == 'staging'):
            command = 'gcloud config get-value account'
            user_email = subprocess.check_output(command, shell=True).decode(sys.stdout.encoding)
            self.user_email = user_email.split("@")[0]

        self.bucket_name = stage_config['bucket_name']
        self.project_id = stage_config['project_id']
        self.project_name = stage_config['project_name']
        self.region = stage_config['region']
        self.zone = self.region + '-b'  # TODO: Check if we can configure this as well
        self.template_id = stage_config['template_id']
        self.unique_iteration_id = stage_config['unique_iteration_id']
        self.cluster_name = stage_config['cluster_conf']['managed_cluster']['cluster_name']
        self.cluster_duration = stage_config['cluster_duration_sec']
        self.managed_cluster = stage_config['cluster_conf']['managed_cluster']

        if self.unique_iteration_id:
            self.folder_path = self.project_name + '/' + self.user_email + \
                               '/unique_iteration_id_' + self.unique_iteration_id
        else:
            self.folder_path = self.project_name + '/' + self.user_email + '/main'

        self.bucket_path = f'gs://{self.bucket_name}/{self.folder_path}'
        self.project_path = 'projects/{project_id}/regions/{region}'.format(project_id=self.project_id,
                                                                            region=self.region)

    def create_dataproc_cluster(self):
        """! Create a GCP cluster on DataProc based on configurations from batch_config.json
        """

        # Defined used paths
        main_project_path = os.path.abspath(os.getcwd())

        # Extract pip packages for the cluster
        pip_packages = ''
        with open(main_project_path + '/requirements_docker.txt') as f:
            required = f.read().splitlines()
            required = [x for x in required if "#" not in x]
            pip_packages = ' '.join(required)

        client = ClusterControllerClient(
            client_options={"api_endpoint": f"{self.region}-dataproc.googleapis.com:443"}
        )

        # Instantiate a cluster
        cluster = Cluster()
        cluster.cluster_name = self.cluster_name
        if self.managed_cluster:
            self.config['cluster_conf']['managed_cluster']['config']['gce_cluster_config']['metadata'][
                'PIP_PACKAGES'] = pip_packages
        managed_cluster_config = self.config['cluster_conf']['managed_cluster']['config']
        cluster_config = ClusterConfig(managed_cluster_config)

        cluster_duration = Duration(seconds=self.cluster_duration)
        cluster_config.lifecycle_config = {
            'idle_delete_ttl': cluster_duration
        }

        cluster.config = cluster_config

        # Getting cluster if it exists
        exist_cluster = None
        try:
            exist_cluster = client.get_cluster(
                project_id=self.project_id,
                region=self.region,
                cluster_name=self.cluster_name,
            )
            print(f'Cluster {self.cluster_name} retrieved successfully')

        except Exception as e:
            # A "404", not found exception, is expected here and should only be noted
            if "404" in str(e):
                print(f"Cluster {self.cluster_name} not found. Creating...")
            else:
                print(f'Error getting cluster: {e}')
                raise e

        # Updating the cluster if it exists
        if exist_cluster:
            try:
                updated_config_list = [
                    'config.worker_config.num_instances',
                    'config.secondary_worker_config.num_instances',
                    'labels'
                ]
                client.update_cluster(
                    project_id=self.project_id,
                    region=self.region,
                    cluster_name=self.cluster_name,
                    cluster=cluster,
                    update_mask={
                        "paths": updated_config_list
                    }
                )
                print(f'cluster {self.cluster_name} was updated successfully')

            except Exception as e:
                print(f'error updating cluster: {e}')

        # If no existing cluster was found, create it
        else:
            try:
                client.create_cluster(
                    project_id=self.project_id,
                    region=self.region,
                    cluster=cluster
                )
                print(f'Cluster {self.cluster_name} created successfully')

            except Exception as e:
                print(f'Error creating cluster: {e}')

    def get_client_check_cluster(self) -> (bool, ClusterControllerClient):
        """! Set up a client and check for an existing cluster
            RETURNS:
                Cluster: If a cluster was found, return its class
                ClusterControllerClient: Client for controlling the cluster
        """

        # Set up a client
        client = ClusterControllerClient(
            client_options={"api_endpoint": f"{self.region}-dataproc.googleapis.com:443"}
        )

        # Getting cluster if it exists
        cluster = None
        cluster_exists = False
        try:
            cluster = client.get_cluster(
                project_id=self.project_id,
                region=self.region,
                cluster_name=self.cluster_name,
            )
            print(f'Cluster {self.cluster_name} retrieved successfully')

        except Exception as e:
            # A "404", not found exception, is expected here and should only be noted
            if "404" in str(e):
                print(f"Cluster {self.cluster_name} not found.")
            else:
                print(f'Error getting cluster: {e}')
                raise e
        if cluster:
            cluster_exists = True
        return client, cluster_exists

    def start_dataproc_cluster(self):
        """! Start an existing GCP cluster on DataProc based on configurations from batch_config.json
        """
        client, existing_cluster = self.get_client_check_cluster()

        if existing_cluster:
            try:
                operation = client.start_cluster(
                    request={"project_id": self.project_id,
                             "region": self.region,
                             "cluster_name": self.cluster_name}
                )
                response = operation.result()
                print(f'Started cluster {self.cluster_name}. Result: {response} ')

            except Exception as e:
                print(f'Failed starting cluster {self.cluster_name}')

    def stop_dataproc_cluster(self):
        """! Stop an existing GCP cluster on DataProc based on configurations from batch_config.json
        """

        client, existing_cluster = self.get_client_check_cluster()

        if existing_cluster:
            try:
                operation = client.stop_cluster(
                    request={"project_id": self.project_id,
                             "region": self.region,
                             "cluster_name": self.cluster_name}
                )
                response = operation.result()
                print(f'Stopped cluster {self.cluster_name}. Result: {response} ')

            except Exception as e:
                print(f'Failed stopping cluster {self.cluster_name}')

    def delete_dataproc_cluster(self):
        """!Delete an existing DataProc cluster according to project specific configuration.
        Since we can't update certain parameters on an existing cluster, we need the option
        to delete it
        """

        client, existing_cluster = self.get_client_check_cluster()

        if existing_cluster:
            try:
                operation = client.delete_cluster(
                    project_id=self.project_id,
                    region=self.region,
                    cluster_name=self.cluster_name,
                )
                response = operation.result()
                print(f'Deleted cluster {self.cluster_name}. Result: {response} ')

            except Exception as e:
                print(f'Failed deleting cluster {self.cluster_name}')

    def upload_files_to_bucket(self, upload_type: str):
        """! Upload jars, stages and wheel file to a given bucket
                ARGS:
                    upload_type: String reflecting which file type we wish to upload to bucket (jars, stages, whl, all)
        """
        parent_path = os.path.abspath(os.path.join(os.getcwd(), 'batch_files'))

        if upload_type == 'all' or upload_type == 'jars':
            self.upload_jars_to_bucket(parent_path)
        if upload_type == 'all' or upload_type == 'stages':
            self.upload_stages_main_to_bucket(parent_path)
        if upload_type == 'all' or upload_type == 'whl':
            self.upload_whl_to_bucket(parent_path)

    def upload_jars_to_bucket(self, root_dir):
        """! Upload relevant jars to a given bucket
            ARGS:
                root_dir: Directory in which we search for the jars directory
        """

        for path, subdirs, files in os.walk(root_dir + '/jars'):
            for name in files:
                path_local = os.path.join(path, name)
                bucket_blob_path = path_local[path_local.index(self.project_name) +
                                              len(self.project_name):].replace('\\', '/')
                # blob = bucket.blob(folder_path + bucket_blob_path)
                # blob.upload_from_filename(path_local)
                self.gsutils_upload_one_file(path_local, bucket_blob_path)
            # print(path_local)
            # print(bucket_blob_path)
        print('uploaded jars main successfully')

    def upload_stages_main_to_bucket(self, root_dir):
        """! Upload relevant stages to a given bucket
            ARGS:
                root_dir: Directory in which we search for the stages directory
        """

        for path, subdirs, files in os.walk(root_dir + '/stages'):
            for name in files:
                if name == 'main.py':
                    path_local = os.path.join(path, name)
                    bucket_blob_path = path_local[path_local.index(self.project_name) +
                                                  len(self.project_name):].replace('\\', '/')
                    # blob = bucket.blob(folder_path + bucket_blob_path)
                    # blob.upload_from_filename(path_local)
                    self.gsutils_upload_one_file(path_local, bucket_blob_path)
                # print(path_local)
                # print(bucket_blob_path)
                # print(name)
        print('uploaded stages main successfully')

    def upload_whl_to_bucket(self, root_dir):
        """! Upload relevant wheel file to a given bucket
            ARGS:
                root_dir: Directory in which we search for the wheel archive
        """

        storage_client = storage.Client(project=self.project_id)
        bucket = storage_client.get_bucket(self.bucket_name)
        parent_path = os.path.abspath(os.path.join(root_dir, ".."))
        dist = parent_path + '/dist'
        onlyfiles = [os.path.join(dist, f) for f in os.listdir(dist) if os.path.isfile(os.path.join(dist, f))]
        for path_local in onlyfiles:
            bucket_blob_path = path_local[path_local.index(self.project_name) +
                                          len(self.project_name):].replace('\\', '/')
            blob = bucket.blob(self.folder_path + bucket_blob_path)
            blob.chunk_size = 1024 * 1024
            blob.upload_from_filename(path_local)
        # print(path_local)
        # print(project_bucket_folder + bucket_blob_path)
        # print('onlyfiles', onlyfiles)
        # for path, subdirs, files in os.walk(parent_path + '/dist'):
        # 	files = [file for file in files if file.endswith(('.zip', '.whl', '.egg'))]
        # 	for name in files:
        # 		path_local = os.path.join(path, name)
        # 		bucket_blob_path = path_local[path_local.index(orig_project_name) +
        # 		len(orig_project_name):].replace('\\', '/')
        # 		blob = bucket.blob(project_bucket_folder + bucket_blob_path)
        # 		# blob.chunk_size = 1024 * 1024
        # 		# blob.upload_from_filename(path_local)
        # 		# print(path_local)
        # 		# print(bucket_blob_path)
        # 		# print(bucket_blob_path2)
        # 		# print(name)
        # 		# print(path)
        print('uploaded package whl successfully')

    def gsutils_upload_one_file(self, local_path, target_path):
        """! Upload relevant wheel file to a given bucket
            ARGS:
                local_path: Directory from which we copy the data
                target_path: Directory on the remote location to which we upload the data
        """

        command = '''gsutil -o GSUtil:parallel_composite_upload_threshold=150M -m cp \
        "''' + local_path + '''" \
        "''' + self.bucket_path + target_path + '''" \
        '''

        process = subprocess.call(command, shell=True)




from typing import Union, List

import logging
import sys
from pythonjsonlogger import jsonlogger

##  Logger setup
root = logging.getLogger()
root.setLevel(logging.INFO)
## Redirect logging to sys.stdout
handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.INFO)
## Assign logger a json formatter to output log data as json.
formatter = jsonlogger.JsonFormatter()
handler.setFormatter(formatter)
root.addHandler(handler)

logger = logging.getLogger(__name__)

import os
import json
import uvicorn
from fastapi import FastAPI, Depends, Header, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from starlette.requests import Request

from pipeline.pipeline import generatedProjectNamePipeline
from pipeline.schema.inputs import generatedProjectNameInputs
from pipeline.schema.outputs import generatedProjectNameOutputs
from tester.test_schema.test_input import TestInputRequest
from tester.test_schema.test_output import TestOutputResponse
from tester.general_tester import generatedProjectNameGeneralTester


async def get_token_header(x_token: str = Header(...)):
    if x_token != "fake-super-secret-token": # todo: please replace with realistic token
        raise HTTPException(status_code=400, detail="X-Token header invalid")


async def verify_env():
    # Don't allow usage of an endpoint on production environment
    if os.environ.get('SPRING_PROFILES_ACTIVE') == 'production':
        raise HTTPException(status_code=404, detail="Endpoint not available")


def load_allowed_cors_origins():
    allowed_origins = []
    cors_origin_list_path = 'server/cors_allowed_origins.json'
    if os.path.exists(cors_origin_list_path):
        with open(cors_origin_list_path) as f:
            allowed_origins = json.load(f)

    return allowed_origins

## Instantiate FastAPI()
app = FastAPI()
## Allowed cors origins
origins = load_allowed_cors_origins()
## Methods allowed, for FastAPI()
methods = ["*"]
## Headers allowed, for FastAPI()
headers = ["*"]
## Credentials required, bool.
credentials = True
##
# @var allow_origins
# Cors allowed origins.
# @var allow_credentials
# Allowed credentials by FastAPI()
# @var allow_methods
# Allowed methods by FastAPI()
# @var allow_headers
# Allowed headers by FastAPI()
app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=credentials,
    allow_methods=methods,
    allow_headers=headers,
)
## Initialized a pipeline for /predict /parse endpoints.
pipeline = generatedProjectNamePipeline()
# from server.pool import InstancePool
# pipelines = [generatedProjectNamePipeline() for i in range(3)]
# app.pipelines = InstancePool(pipelines)


@app.post('/predict', dependencies=[Depends(get_token_header)])
def predict(body: generatedProjectNameInputs, request: Request) -> List[generatedProjectNameOutputs]:
    """! Predict api endpoint, is designed to support predict projects such as
    scoops classification, valid user email etc.

    Use by post request, for example:

        headers = {"Content-Type": "application/json; charset=UTF-8", **global_headers}
        url = http://localhost:8080/predict
        data = {}  # The data to run by model.
        response = requests.post(url, json=data, headers=headers)

    """
    data = body.dict()
    # logger.info({"message": "predict invoked", "data": json.dumps(data)})
    # print('data', data)

    # call model here
    try:
        output: generatedProjectNameOutputs = pipeline.execute(**data)
        logger.info("INFO predict invoked", extra={"input": data, "output": output.dict()})
    except Exception as ex:
        logger.exception(msg=data, exc_info=True)
        output = {'error': {'request': ex}}
        logger.error("ERROR predict invoked", extra={"input": data, "output": output})
    # call model here
    return output


@app.post('/parse', dependencies=[Depends(get_token_header)])
def parse(body: generatedProjectNameInputs, request: Request) -> List[generatedProjectNameOutputs]:
    """! Parse api endpoint, is designed to support parse projects such as trex, bio parser, sig parser etc.

    Use by post request, for example:

        headers = {"Content-Type": "application/json; charset=UTF-8", **global_headers}
        url = http://localhost:8080/parse
        data = {}  # The data to run by model.
        response = requests.post(url, json=data, headers=headers)

    """
    data = body.dict()
    # logger.info({"message": "parse invoked", "data": json.dumps(data)})
    # print('data', data)

    # call model here
    try:
        output: generatedProjectNameOutputs = pipeline.execute(**data)
        logger.info("INFO parse invoked", extra={"input": data, "output": output.dict()})
    except Exception as ex:
        logger.exception(msg=data, exc_info=True)
        output = {'error': {'request': ex}}
        logger.error("ERROR parse invoked", extra={"input": data, "output": output})
    # call model here
    return output


## Fetch the proper tester for the project
model_tester = generatedProjectNameGeneralTester()


@app.post('/test', dependencies=[Depends(get_token_header), Depends(verify_env)], include_in_schema=False)
def test(input_request: TestInputRequest) -> List[TestOutputResponse]:
    """! This is the main endpoint for testing.

        Args:
            input_request: Request as received from the DSP
        Returns:
            A list of responses, to be sent back to the DSP

    Every project starts with a default mock response service, in order to enable DSP integration (when the working
    testing functions are ready and implemented in the tester/general_tester.py),
    replace the mock service with the real one, in practice just remark the following line:

        response = model_tester.create_mock_response(input_request)

    and remove remark quotes from the following block of code:

        model_tester.save_meta_data(input_request)
        response = []
        # Until this model will support batches, work sequentially. Once it does, pass the full list of rows.
        for request_row in input_request.rows:
            input_list = [request_row]
            tester_output = model_tester.test_batch(input_list)
            response.append(tester_output[0])

    """
    response = model_tester.create_mock_response(input_request)

    """    
    model_tester.save_meta_data(input_request)
    response = []
    # Until this model will support batches, work sequentially. Once it does, pass the full list of rows.
    for request_row in input_request.rows:
        input_list = [request_row]
        tester_output = model_tester.test_batch(input_list)
        response.append(tester_output[0])
    """
    return response


@app.get("/livenessprobe")
def liveness_probe():
    """! Endpoint used to test server liveness

        Usage:
            http://localhost:8080/livenessprobe

        Returns:
            server response: {"alive": True} - On success

    """
    return {"alive": True}


@app.get("/readinessprobe")
def readiness_probe():
    """! Endpoint used to test server readiness

        Usage:
            http://localhost:8080/readinessprobe

        Returns:
            server response: {"ready": True} - On success

        """
    return {"ready": True}


if __name__ == '__main__':
    # Runs the server
    ##
    # @var host
    # IP Address
    # @var port
    # Port number
    uvicorn.run(app, host='0.0.0.0', port=8080)




"""! @brief ZIDS_Stage base class for the batch stage class."""

import json
import sys
from pyspark.sql import SparkSession
from typing import List, Any

##
# @file
# @brief Defines stage base class.
class ZIDS_Stage():
    def __init__(self, stage_config):
        """! ZIDS_Stage initializer.
        Loads config file.

            Args:
                stage_config : Configuration dictionary, loaded from configuration file.

        """
        self.bucket_name = stage_config['bucket_name']
        self.project_id = stage_config['project_id']
        self.project_name = stage_config['project_name']
        self.region = stage_config['region']
        self.unique_iteration_id = stage_config['unique_iteration_id']
        if self.unique_iteration_id:
            print(f'unique_iteration_id: {self.unique_iteration_id}')
            self.project_name = self.project_name + '/unique_iteration_id_' + self.unique_iteration_id
        else:
            self.project_name = self.project_name + '/main'

        self.bucket_path = f'gs://{self.bucket_name}/{self.project_name}'
        self.start_date = ""

    def update_start_date(self, stage_start_date):
        """! Update start date
            ARGS:
                stage_start_date: String, containing the starting date, received from Airflow
        """
        print(f'Setting start date: {stage_start_date}')
        self.start_date = stage_start_date

    def main(self, **kwargs: Any):
        """! ZIDS_Stage main function.
        This function is the "entrypoint" for the stage, this will run when the job is executed.

            Args:
                **kwargs : Whatever is needed for the stage to run properly.
        """
        raise NotImplementedError

    def load_spark(self) -> SparkSession:   # TODO: Generalize this in order to receive config options.
        """! Basic loading of a spark session.

            Returns:
                A basic spark session
        """
        spark = SparkSession.builder \
            .appName(self.project_name) \
            .config('spark.ui.showConsoleProgress', True) \
            .config("spark.sql.parquet.compression.codec", "gzip") \
            .config('spark.jars', 'gs://spark-lib/bigquery/spark-bigquery-latest_2.12.jar') \
            .getOrCreate()

        print(f"Spark ID: {spark.sparkContext.applicationId}")
        return spark




